"""
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
Helper functions for the file selection GUI.
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
"""

#!/usr/bin/python
# -*- coding: utf-8 -*-
#
# License:
#
#
# Revit Batch Processor Sample Code
#
# BSD License
# Copyright 2023, Jan Christel
# All rights reserved.

# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

# - Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
# - Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
# - Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
#
# This software is provided by the copyright holder "as is" and any express or implied warranties, including, but not limited to, the implied warranties of merchantability and fitness for a particular purpose are disclaimed.
# In no event shall the copyright holder be liable for any direct, indirect, incidental, special, exemplary, or consequential damages (including, but not limited to, procurement of substitute goods or services; loss of use, data, or profits;
# or business interruption) however caused and on any theory of liability, whether in contract, strict liability, or tort (including negligence or otherwise) arising in any way out of the use of this software, even if advised of the possibility of such damage.
#
#
#

# this sample shows how to write out a number of task files using bucket distribution by file size

import clr
import System
import os


# import file item class
from duHast.UI.Objects import file_item as fi

# import workloader utils
from duHast.UI import workloader as wl

# custom result class
from duHast.Utilities.Objects import result as res
from duHast.Utilities.files_io import (
    is_back_up_file,
    get_file_size as files_io_get_file_size,
    FILE_SIZE_IN_KB,
)
from duHast.Utilities.files_get import (
    get_files,
    get_files_from_directory_walker_with_filters_simple,
)
from duHast.Utilities.files_csv import get_first_row_in_csv_file, read_csv_file
from duHast.Revit.BIM360.util_bim_360 import get_bim_360_file_data


# -------------
# my code here:
# -------------


def get_revit_files_for_processing(location, include_sub_dirs, file_extension):
    """
    Extracts file data from varies sources:
        - bim 360 task text file
        - files on a local server directory
        - files on a local server directory and it's subdirectories
        - local server based files provided through a task text file

    :param location: Can either be a directory, or a fully qualified file path.
    :type location: str
    :param includeSubDirs: If true and location is a directory, subdirectories will be included in file search.
    :type includeSubDirs: bool
    :param fileExtension: File type filter in '.ext'
    :type fileExtension: str

    :return: List of MyFileItem objects.
    :rtype: :class:`.MyFileItem`
    """

    revit_files = []
    try:
        if os.path.isfile(location):
            # got a text file...could either be BIM360 data or a file task list
            revit_files = _get_file_data_from_text_file(location)
        elif os.path.isdir(location):
            # check a to search for files is to include sub dirs
            revit_files_unfiltered = []
            if include_sub_dirs:
                # get revit files in input dir and subdirs (with backup files removed already)
                revit_files_unfiltered = get_revit_files_incl_sub_dirs(
                    location, file_extension
                )
            else:
                # get revit files in input dir (with backup files removed already)
                revit_files_unfiltered = get_revit_files(location, file_extension)
            # check for max path violations!
            # The specified path, file name, or both are too long. The fully qualified file name must be less than 260 characters, \
            # and the directory name must be less than 248 characters.
            for revit_file in revit_files_unfiltered:
                # remove any back up files from selection
                if (
                    len(os.path.dirname(os.path.abspath(revit_file.name))) < 248
                    and len(revit_file.name) < 260
                ):
                    revit_files.append(revit_file)
    except Exception as e:
        # return an empty list which will cause this script to abort
        revit_files = []
    return revit_files


def _get_files_from_list_file(file_path_csv):
    """
    Reads server based file data, the fully qualified file path, from a task file list file in csv format.

    :param filePathCSV: The fully qualified file path to the task list file.
    :type filePathCSV: str

    :return: A list of MyFileitem objects. If an exception occured an empty list will be returned.
    :rtype: :class:`.MyFileItem`
    """

    revit_files = []
    try:
        # attempt to read the CSV into rows
        rows_result = read_csv_file(file_path_csv)
        if rows_result.status is False:
            return []
        rows = rows_result.result
        
        # check whether anything came back
        if len(rows) > 0:
            # process rows
            for row_data in rows:
                if len(row_data) > 0:
                    file_size = files_io_get_file_size(row_data[0], FILE_SIZE_IN_KB)
                    dummy = fi.MyFileItem(row_data[0], file_size)
                    revit_files.append(dummy)
    except Exception as e:
        # return an empty list which will cause this script to abort
        revit_files = []
    return revit_files


def get_files_from_csv_list_file(file_path_csv, file_extension):
    """
    Reads server based file data, the fully qualified file path, from a task file list file in csv format.

    :param filePathCSV: The fully qualified file path to the task list file.
    :type filePathCSV: str
    :param fileExtension: The file extension filter in format '.ext'
    :type fileExtension: str
    :return: A list of MyFileitem objects. If an exception occured an empty list will be returned.
    :rtype: :class:`.MyFileItem`
    """

    revit_files = _get_files_from_list_file(file_path_csv)
    # filter out files with wrong extension
    revit_files = [f for f in revit_files if f.name.lower().endswith(file_extension)]
    return revit_files


def _get_file_data_from_text_file(file_path):
    """
    Reads a file server based task list file. This file can either be a BIM360 task list file or \
        a task list file containing file server based file path in a single column.

    :param filePath: The fully qualified file path to the task list file.
    :type filePath: str
    :return: A list of MyFileitem objects.
    :rtype: :class:`.MyFileItem`
    """

    files = []

    # if file is empty an empty list will be returned
    # also need to check whether this is a csv file...
    if file_path.lower().endswith(".csv"):
        # list of entries in first row
        row_result = get_first_row_in_csv_file(file_path)
        if row_result.status:
            row = row_result.result
    else:
        row = get_first_row_in_file_no_strip(file_path)
        # make sure we get a list of entries
        if row is not None:
            row = row.split("\t")
    if row is not None:
        # bim 360 or autodesk construction cloud files have at least 3 entries
        if len(row) > 2:
            files = get_bim_360_file_data(file_path)
        else:
            files = _get_files_from_list_file(file_path)
    return files


def get_revit_files(directory, file_extension):
    """
    Returns files in a given directory and of a given file extension.

    :param directory: The fully qualified directory path.
    :type directory: str
    :param file_extension: The file extension filter in format '.ext'
    :type file_extension: str

    :return: List of file items
    :rtype: [:class:`.FileItem`]
    """

    files = []
    list_of_files = get_files(directory, file_extension)
    for f in list_of_files:
        # check if this is a back up file
        if is_back_up_file(f) == False:
            # Get size and add to list of files.
            size = os.path.getsize(f)
            files.append(fi.MyFileItem(f, size))
    return files


def get_revit_files_incl_sub_dirs(directory, file_extension):
    """
    Returns files in a given directory and its sub directories of a given file extension.

    :param directory: The fully qualified directory path.
    :type directory: str
    :param file_extension: The file extension filter in format '.ext'
    :type file_extension: str

    :return: List of file items
    :rtype: [:class:`.FileItem`]
    """

    files = []
    # Get the list of all files in directory tree at given path
    list_of_files = get_files_from_directory_walker_with_filters_simple(
        directory, file_extension
    )

    for f in list_of_files:
        # check if this is a back up file,
        if is_back_up_file(f) == False:
            # Get size and add to list of files.
            size = os.path.getsize(f)
            files.append(fi.MyFileItem(f, size))
    return files


def get_file_size(item):
    """
    Helper used to define workload size (same as file size)

    :param item: A file item object instance.
    :type item: :class:`.FileItem`

    :return: The file size.
    :rtype: int
    """

    return item.size


def bucket_to_task_list_file_system(item):
    """
    Default task list content for files on a file server location.

    :param item: A file item object instance.
    :type item: :class:`.FileItem`

    :return: The item name.(fully qualified file path)
    :rtype: str
    """

    return item.name


def bucket_to_task_list_bim_360(item):
    """
    Default task list content for files on a BIM 360 cloud drive.

    :param item: A file item object instance.
    :type item: :class:`.FileItem`

    :return: The revit version, project guid, file guid separated by a space ' '
    :rtype: str
    """

    return " ".join(
        [item.bim_360_revit_version, item.bim_360_project_guid, item.bim_360_file_guid]
    )


def write_revit_task_file(file_name, bucket, get_data=bucket_to_task_list_file_system):
    """
    Writes out a task list file.

    :param file_name: Fully qualified file path of the task file name including extension.
    :type file_name: str
    :param bucket: Workload bucket object instance.
    :type bucket: :class:`.WorkloadBucket`
    :param get_data: Returns data from file item object to be written to file, defaults to BucketToTaskListFileSystem
    :type get_data: func(:class:`.FileItem`) -> str, optional

    :return:
        Result class instance.

        - Write file status (bool) returned in result.status. False if an exception occurred, otherwise True.
        - Result.message property contains fully qualified file path to task list file.

        On exception:

        - .status (bool) will be False.
        - .message will contain the exception message.

    :rtype: :class:`.Result`
    """

    return_value = res.Result()
    try:
        f = open(file_name, "w")
        row_counter = 0
        for p in bucket.items:
            data = get_data(p)
            # check whether first row
            if row_counter != 0:
                # if not first row add line feed character before data written
                data = "\n" + data
            else:
                row_counter += 1
            # this looks horrible: but:
            # string.encode('utf-8') returns a byte string b''
            # so it needs to be converted to an actual string again
            f.write(data.encode("utf-8").decode("utf-8"))
        f.close()
        return_value.append_message("wrote task list: {} [TRUE]".format(file_name))
        return_value.result.append(file_name)
    except Exception as e:
        return_value.update_sep(
            False,
            "Failed to write task list: {} with exception {}".format(file_name, e),
        )
    return return_value


def write_file_list(
    directory_path,
    file_extension,
    task_list_directory,
    task_files_number,
    file_getter,
    file_data_processor=bucket_to_task_list_file_system,
):
    """
    Writes out all task list(s) to file(s).

    :param directory_path: Fully qualified directory path containing files to be added to task lists.
    :type directory_path: str
    :param file_extension: A file extension filter in format '.ext'
    :type file_extension: str
    :param task_list_directory: The fully qualified directory path where the task files will be written .
    :type task_list_directory: str
    :param task_files_number: The number of task files to be written.
    :type task_files_number: int
    :param file_getter: Function accepting a directory and file extension filter and returns file items from directory.
    :type file_getter: func(str, str) -> :class:`.FileItem`
    :param file_data_processor: Function processing file item and returns a string to be written to task list file, defaults to BucketToTaskListFileSystem
    :type file_data_processor: func(:class:`.FileItem`) -> str, optional

    :return:
        Result class instance.

        - Write file status (bool) returned in result.status. False if an exception occurred, otherwise True.
        - Result.message property contains fully qualified file path for each task list file.

        On exception:

        - .status (bool) will be False.
        - .message will contain the exception message.

    :rtype: :class:`.Result`
    """

    return_value = res.Result()
    return_value.status = True
    # get revit files in input dir
    revit_files = file_getter(directory_path, file_extension)
    # build bucket list
    buckets = wl.distribute_workload(task_files_number, revit_files, get_file_size)
    try:
        # write out file lists
        counter = 0
        for bucket in buckets:
            file_name = os.path.join(
                task_list_directory, "Tasklist_" + str(counter) + ".txt"
            )
            status_write = write_revit_task_file(file_name, bucket, file_data_processor)
            return_value.update(status_write)
            counter += 1
        return_value.append_message("Finished writing out task files")
    except Exception as e:
        return_value.update_sep(False, "Failed to save file list! " + str(e))
    return return_value


def get_task_file_name(task_list_directory, counter):
    """
    Builds a fully qualified task file path.

    :param taskListDirectory: The fully qualified directory where the task file is (to be) located
    :type taskListDirectory: str
    :param counter: A counter to be added to the task file name
    :type counter: int

    :return: A fully qualified task file path
    :rtype: str
    """

    file_name = os.path.join(task_list_directory, "Tasklist_" + str(counter) + ".txt")
    return file_name


def write_empty_task_list(file_name):
    """
    Writes out an empty task list.

    :param fileName: Fully qualified file path of the task file name including extension.
    :type fileName: str

    :return:
        Result class instance.

        - Write file status (bool) returned in result.status. False if an exception occurred, otherwise True.
        - Result.message property contains fully qualified file path to empty task list file.

        On exception:

        - .status (bool) will be False.
        - .message will contain the exception message.

    :rtype: :class:`.Result`
    """

    return_value = res.Result()
    try:
        f = open(file_name, "w")
        f.close()
        return_value.append_message(
            "wrote empty task list: {} [TRUE]".format(file_name)
        )
    except Exception as e:
        return_value.update_sep(
            False,
            "Failed to write empty task list: {} with exception: {}".format(
                file_name, e
            ),
        )
    return return_value



"""
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
This module contains a number of helper functions relating to BIM360.
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
"""
#
# License:
#
#
# Revit Batch Processor Sample Code
#
# BSD License
# Copyright 2023, Jan Christel
# All rights reserved.

# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

# - Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
# - Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
# - Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
#
# This software is provided by the copyright holder "as is" and any express or implied warranties, including, but not limited to, the implied warranties of merchantability and fitness for a particular purpose are disclaimed.
# In no event shall the copyright holder be liable for any direct, indirect, incidental, special, exemplary, or consequential damages (including, but not limited to, procurement of substitute goods or services; loss of use, data, or profits;
# or business interruption) however caused and on any theory of liability, whether in contract, strict liability, or tort (including negligence or otherwise) arising in any way out of the use of this software, even if advised of the possibility of such damage.
#
#
#

import System
import clr


from duHast.Utilities.files_csv import read_csv_file
from duHast.UI.Objects import file_item as fi


def get_bim_360_revit_files_from_file_list(file_path_csv, extension):
    """
    Method reading BIM360 file data from a CSV file.

    :param file_path_csv: Fully qualified file path to CSV to be read.
    :type file_path_csv: str
    :param extension: Is an empty place holder...this method is past into another where it expects 2 arguments...
    :type extension: TODO
    :raise: On exception an empty list is returned.

    :return: List in format [Revit version, Project GUID, File GUID, file size in MB, file name]
    :rtype: list of str
    """

    files = get_bim_360_file_data(file_path_csv)
    return files


def get_bim_360_file_data(file_path_csv):
    """
    Reads a csv file containing BIM 360 file data into list of FileItem instances.

    :param file_path_csv: Fully qualified file path to CSV to be read.
    :type file_path_csv: str
    :return: a list of FileItem instances representing BIM360 file data.
        Will return an empty list of an exception occurred.
    :rtype: list of SampleCodeBatchProcessor.FileItem
    """
    revit_files = []
    try:
        # attempt to read the CSV into rows
        rows_result = read_csv_file(file_path_csv)
        if(rows_result.status is False):
            return []
        rows = rows_result.result
        
        # check whether anything came back
        if len(rows) > 0:
            # process rows
            for row in rows:
                dummy = process_bim_360_file_data_row(row)
                # check whether row got processed ok
                if dummy is not None:
                    revit_files.append(dummy)
    except Exception as e:
        print("An exception occurred during BIM360 row processing! {}".format(e))
        # return an empty list which will cause this script to abort
        revit_files = []
    return revit_files


def process_bim_360_file_data_row(row_data):
    """
    Reads a list of str into file item class object

    :param row_data: the list containing the file data
    :type row_data: list of str in format [Revit version, Project GUID, File GUID, file size in MB, file name]

    :return: Initialized file item instance, None if row is not the right length
    :rtype: SampleCodeBatchProcessor.FileItem
    """

    # check whether we have the right number of columns
    if len(row_data) == 5:
        dummy = fi.MyFileItem(
            name=row_data[4],
            size=int(row_data[3]),
            bim360_project_guid=row_data[1],
            bim360_file_guid=row_data[2],
            bim360_revit_version=row_data[0],
        )
        return dummy
    else:
        return None



import logging


class FilterFile(logging.Filter):
    """
    Filter to stop logs to file handler.

    To use change your log call to something like:
    my_logger.info("This is a console only output", extra={"block": "file"})
    """

    def filter(self, record):
        if "block" in record.__dict__.keys():
            if record.block == "file":
                return False
        return True


class FilterConsole(logging.Filter):
    """
    Filter to stop logs to console handler.

    To use change your log call to something like:
    my_logger.info("This is a file only output", extra={"block": "console"})
    """

    def filter(self, record):
        if "block" in record.__dict__.keys():
            if record.block == "console":
                return False
        return True



"""
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
This module contains utility function(s) for level reports. 
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
"""
#
# License:
#
#
# Revit Batch Processor Sample Code
#
# BSD License
# Copyright 2023, Jan Christel
# All rights reserved.

# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

# - Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
# - Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
# - Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
#
# This software is provided by the copyright holder "as is" and any express or implied warranties, including, but not limited to, the implied warranties of merchantability and fitness for a particular purpose are disclaimed.
# In no event shall the copyright holder be liable for any direct, indirect, incidental, special, exemplary, or consequential damages (including, but not limited to, procurement of substitute goods or services; loss of use, data, or profits;
# or business interruption) however caused and on any theory of liability, whether in contract, strict liability, or tort (including negligence or otherwise) arising in any way out of the use of this software, even if advised of the possibility of such damage.
#
#
#


import Autodesk.Revit.DB as rdb

from duHast.Revit.Common import worksets as rWork
from duHast.Utilities import utility as util
from duHast.Utilities.unit_conversion import convert_imperial_feet_to_metric_mm


def get_level_report_data(doc, revitFilePath):
    """
    Gets level data ready for being printed to file.
    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document
    :param revitFilePath: The file hostname, which is added to data returned.
    :type revitFilePath: str
    :return: list of list of revit level properties.
    :rtype: list of list of str
    """

    data = []
    for p in rdb.FilteredElementCollector(doc).OfClass(rdb.Level):
        data.append(
            [
                revitFilePath,
                str(p.Id.IntegerValue),
                util.encode_ascii(p.Name),
                util.encode_ascii(
                    rWork.get_workset_name_by_id(doc, p.WorksetId.IntegerValue)
                ),
                str(convert_imperial_feet_to_metric_mm(p.Elevation)),
            ]
        )
    return data



"""
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
area separation lines solver class.
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

Currently address' area separation lines overlapping with other area separation lines only.
"""
#
# License:
#
#
# Revit Batch Processor Sample Code
#
# BSD License
# Copyright 2023, Jan Christel
# All rights reserved.

# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

# - Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
# - Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
# - Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
#
# This software is provided by the copyright holder "as is" and any express or implied warranties, including, but not limited to, the implied warranties of merchantability and fitness for a particular purpose are disclaimed.
# In no event shall the copyright holder be liable for any direct, indirect, incidental, special, exemplary, or consequential damages (including, but not limited to, procurement of substitute goods or services; loss of use, data, or profits;
# or business interruption) however caused and on any theory of liability, whether in contract, strict liability, or tort (including negligence or otherwise) arising in any way out of the use of this software, even if advised of the possibility of such damage.
#
#
#

from duHast.Utilities.Objects import result as res
from duHast.Utilities.Objects import base
from duHast.Revit.Warnings.Utility.curves_util import (
    get_curves_from_failure_messages,
    check_curves_overlaps,
    delete_curves,
    modify_curves_by_lengthening,
    modify_curves_by_shortening,
)
from duHast.Revit.Common.transaction import in_transaction
from duHast.Revit.Warnings.warning_guids import AREA_SEPARATION_LINES_OVERLAP


class RevitWarningsSolverAreaSepLinesOverlap(base.Base):
    """
    A solver that addresses overlapping area separation lines in a Revit model.
    It can delete or modify these lines to resolve the overlap issue.

    Example Usage:
    solver = RevitWarningsSolverAreaSepLinesOverlap()
    result = solver.solve_warnings(doc, warnings)
    print(result.status)
    print(result.message)

    Main functionalities:
    - Solving overlapping area separation lines by deleting or modifying them.
    - Ignoring area separation lines that overlap when they are in group instances.
    """

    def __init__(
        self,
        solve_by_lengthening_curves=True,
        group_id=-1,
        transaction_manager=in_transaction,
        callback=None,
    ):
        """
        Class constructor.

        :param solve_by_lengthening_curves: Whether to solve the overlapping curves by lengthening or shortening. Defaults to True.
        :type solve_by_lengthening_curves: bool, optional
        :param group_id: The group ID that the curves need to belong to for processing. Defaults to -1.
        :type group_id: int, optional
        :param transaction_manager: The transaction manager function used to perform the modifications. Defaults to in_transaction.
        :type transaction_manager: function, optional
        :param callback: An optional callback function.
        :type callback: function, optional
        """

        # ini super class to allow multi inheritance in children!
        super(RevitWarningsSolverAreaSepLinesOverlap, self).__init__()
        self.filter_name = "Area separation lines overlap."
        self.solve_by_lengthening_curves = solve_by_lengthening_curves
        self.transaction_manager = transaction_manager
        self.group_id = group_id
        self.callback = callback

    # --------------------------- area separation lines ---------------------------
    #: guid identifying this specific warning
    GUID = AREA_SEPARATION_LINES_OVERLAP

    def solve_warnings(self, doc, warnings):
        """
        Solver addressing overlapping area separation lines.

        - will ignore area separation lines overlapping when in different group instances.

        :param doc: Current Revit model document.
        :type doc: Autodesk.Revit.DB.Document
        :param warnings: List of warnings to be solved.
        :type warnings: Autodesk.Revit.DB.FailureMessage

        :return:
            Result class instance.

            - .status True if all area separation lines could be deleted / modified without an exception. Otherwise False.
            - .message will be 'moved tag to area xyz'

        :rtype: :class:`.Result`
        """

        return_value = res.Result()
        if len(warnings) > 0:
            # extract model lines and geometry from failure messages
            curve_storage_sets = get_curves_from_failure_messages(
                doc=doc, failure_messages=warnings, group_id=self.group_id
            )
            # get curves which are completely within other curves
            curves_to_delete = check_curves_overlaps(
                curves=curve_storage_sets, group_id=self.group_id
            )
            # delete those curves
            delete_curves_status = delete_curves(
                doc=doc, curves_to_delete=curves_to_delete, curve_descriptor="area"
            )
            return_value.update(delete_curves_status)
            if self.solve_by_lengthening_curves:
                # attempt to modify overlapping curves by lengthening the longer curve to
                # the length of the shorter curve and thereby completely overlapping the shorter curve
                modify_curves_status = modify_curves_by_lengthening(
                    doc=doc,
                    guid=self.GUID,
                    group_id=self.group_id,
                    transaction_manager=self.transaction_manager,
                    callback=self.callback,
                )
                return_value.update(modify_curves_status)
            else:
                # attempt to modify overlapping curves by shortening the longer curve to
                # the point of overlap with the shorter curve
                modify_curves_status = modify_curves_by_shortening(
                    doc=doc,
                    guid=self.GUID,
                    group_id=self.group_id,
                    transaction_manager=self.transaction_manager,
                    callback=self.callback,
                )
                return_value.update(modify_curves_status)
        else:
            return_value.update_sep(
                True,
                "{}: No warnings of type: area separation lines overlap in model.".format(
                    self.filter_name
                ),
            )
        return return_value



# License:
#
#
# Revit Batch Processor Sample Code
#
# BSD License
# Copyright 2025, Jan Christel
# All rights reserved.

# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

# - Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
# - Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
# - Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
#
# This software is provided by the copyright holder "as is" and any express or implied warranties, including, but not limited to, the implied warranties of merchantability and fitness for a particular purpose are disclaimed.
# In no event shall the copyright holder be liable for any direct, indirect, incidental, special, exemplary, or consequential damages (including, but not limited to, procurement of substitute goods or services; loss of use, data, or profits;
# or business interruption) however caused and on any theory of liability, whether in contract, strict liability, or tort (including negligence or otherwise) arising in any way out of the use of this software, even if advised of the possibility of such damage.
#
#
#

from duHast.Utilities.Objects.result import Result
from duHast.Revit.Family.LibraryCleanUp.Utility.directives_write_to_file import write_duplicate_directives_to_file, write_families_with_missing_group_codes


def verify_copy_directives_are_unique(copy_directives):

    """
    Verify that the copy directives are unique based on their new names.

    :param copy_directives: A list of copy directives to verify.
    :type copy_directives: list[:class:`.FamilyDirectiveCopy`]
    
    :return: True if all copy directives are unique, False otherwise.
    :rtype: bool
    """
    
    seen_names = set()
    duplicate_directive_new_names = []
    duplicate_directives = []

    # find all duplicate new names
    for directive in copy_directives:
        if directive.new_name in seen_names:
            duplicate_directive_new_names.append(directive.new_name)
            continue
        seen_names.add(directive.new_name)
    
    # get all directive with duplicate new names
    for directive in copy_directives:
        if directive.new_name in duplicate_directive_new_names:
            duplicate_directives.append(directive)


    return duplicate_directives

def check_directives(output_path, copy_directives, families_with__missing_group_codes = None):
    """
    Check the copy directives for duplicates and missing group codes.
    :param output_path: The path where the output files will be written.
    :type output_path: str
    :param copy_directives: A list of copy directives to check.
    :type copy_directives: list[:class:`.FamilyDirectiveCopy`]
    :param families_with__missing_group_codes: A list of families with missing group codes.
    :type families_with__missing_group_codes: list[:class:`.FamilyTypeDataStorageManager`]
    
    :return: A Result object containing the status and messages of the checks.
    :rtype: :class:`.Result`
    """

    return_value = Result()
    try:

        # check missing group codes
        if families_with__missing_group_codes is not None and len(families_with__missing_group_codes) > 0:
            return_value.update_sep(
                False,
                "Found families with missing group codes. Please check the output file for details.",
            )
            write_missing_result = write_families_with_missing_group_codes(families_with__missing_group_codes, output_path)
            return_value.append_message(
                write_missing_result.message,
            )

        else:
            return_value.append_message(
                "No families with missing group codes found."
            )

        # check if families with missing group codes are provided
        duplicate_directives = verify_copy_directives_are_unique(copy_directives)

        # check if any duplicate copy directives were found
        if len(duplicate_directives) > 0:

            return_value.update_sep(
                False,
                "Found duplicate copy directives with the same new name. Please check the output file for details.",
            )
            # write duplicate directives to file
            write_duplicate_directives_result = write_duplicate_directives_to_file(duplicate_directives, output_path)

            return_value.append_message(
                write_duplicate_directives_result.message,
            )

        else:
            return_value.append_message(
                "No duplicate copy directives found."
            )

    except Exception as e:
        return_value.update_sep(
            False,
            "Failed to check directives with exception: {}".format(e),
        )
       
    return return_value


# License:
#
#
# Revit Batch Processor Sample Code
#
# BSD License
# Copyright 2025, Jan Christel
# All rights reserved.

# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

# - Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
# - Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
# - Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
#
# This software is provided by the copyright holder "as is" and any express or implied warranties, including, but not limited to, the implied warranties of merchantability and fitness for a particular purpose are disclaimed.
# In no event shall the copyright holder be liable for any direct, indirect, incidental, special, exemplary, or consequential damages (including, but not limited to, procurement of substitute goods or services; loss of use, data, or profits;
# or business interruption) however caused and on any theory of liability, whether in contract, strict liability, or tort (including negligence or otherwise) arising in any way out of the use of this software, even if advised of the possibility of such damage.
#
#
#

import os
import csv

from duHast.Utilities.Objects.result import Result
from duHast.Utilities.files_csv import write_report_data_as_csv

from duHast.Revit.Family.Utility.family_swap_instances_by_type_utils import write_swap_directives_to_file
from duHast.Revit.Family.Utility.family_copy_directive_utils import write_copy_directives_to_file

from duHast.Revit.Family.Data.Objects.family_directive_copy import FamilyDirectiveCopy

from duHast.Revit.Family.LibraryCleanUp.Utility.defaults import (
    SWAP_DIRECTIVE_FILE_NAME,
    DUPLICATE_COPY_DIRECTIVE_FILE_NAME,
    COPY_DIRECTIVE_FILE_NAME,  
    MAINTAIN_TYPES_BY_FAMILY_FILE_NAME, 
    FAMILIES_WITH_MISSING_GROUP_CODES_FILE_NAME
)


def write_maintain_list(maintain_file_list, output_directory):
    """
    Writes a list of maintain directives to a specified file.

    :param maintain_file_list: List of maintain directives to write.
    :type maintain_file_list: list
    :param output_directory: Directory where the file will be written.
    :type output_directory: str

    :return: Result object indicating success or failure.
    :rtype: Result
    """

    return_value = Result()

    try:
        # build the file path for the maintain directives
        file_path = os.path.join(output_directory, MAINTAIN_TYPES_BY_FAMILY_FILE_NAME)

        print("Writing maintain directives to file: {}".format(file_path))

        # write the maintain directives to the file
        return_value = write_report_data_as_csv(
            file_name=file_path,
            header=[],  # No header needed for maintain directives
            data= maintain_file_list, 
            quoting= csv.QUOTE_MINIMAL,)
        
    except Exception as e:
        return_value.update_sep(
            False,
            "Failed to write maintain directives with exception: {}".format(e),
        )
    
    return return_value


def write_swap_directives(swap_directives, output_directory):
    """
    Writes swap directives to a specified file.

    :param swap_directives: List of swap directives to write.
    :type swap_directives: list
    :param output_directory: Directory where the file will be written.
    :type output_directory: str

    :return: Result object indicating success or failure.
    :rtype: Result
    """

    return_value = Result()

    try:
        # build the file path for the swap directives
        file_path = os.path.join(output_directory, SWAP_DIRECTIVE_FILE_NAME)

        print("Writing swap directives to file: {}".format(file_path))

        # write the swap directives to the file
        return_value = write_swap_directives_to_file(
            swap_directives=swap_directives,
            file_path=file_path
        )

    except Exception as e:
        return_value.update_sep(
            False,
            "Failed to write swap directives with exception: {}".format(e),
        )
    
    return return_value


def write_copy_directives(copy_directives, output_directory):
    """
    Writes copy directives to a specified file.

    :param copy_directives: List of copy directives to write.
    :type swap_directives: list
    :param output_directory: Directory where the file will be written.
    :type output_directory: str

    :return: Result object indicating success or failure.
    :rtype: Result
    """

    return_value = Result()

    try:


        # make sure theses are copy directives
        if not all(isinstance(directive, FamilyDirectiveCopy) for directive in copy_directives):
            return_value.update_sep(
                False,
                "All directives must be of type FamilyDirectiveCopy.",
            )
            return return_value

        # build the file path for the copy directives
        file_path = os.path.join(output_directory, COPY_DIRECTIVE_FILE_NAME)

        print("Writing copy directives to file: {}".format(file_path))

        # write the duplicate copy directives to the file
        return_value = write_copy_directives_to_file(
            copy_directives=copy_directives,
            file_path=file_path
        )

    except Exception as e:
        return_value.update_sep(
            False,
            "Failed to write swap directives with exception: {}".format(e),
        )
    
    return return_value

def write_duplicate_directives_to_file(copy_directives, output_directory):
    """
    Writes duplicate directives to a specified file.

    :param copy_directives: List of duplicate directives to write.
    :type copy_directives: list
    :param output_directory: Directory where the file will be written.
    :type output_directory: str

    :return: Result object indicating success or failure.
    :rtype: Result
    """

    return_value = Result()

    try:

        # make sure theses are copy directives
        if not all(isinstance(directive, FamilyDirectiveCopy) for directive in copy_directives):
            return_value.update_sep(
                False,
                "All directives must be of type FamilyDirectiveCopy.",
            )
            return return_value
        
        # build the file path for the duplicate copy directives
        file_path = os.path.join(output_directory, DUPLICATE_COPY_DIRECTIVE_FILE_NAME)

        print("Writing duplicate directives to file: {}".format(file_path))

        # write the duplicate copy directives to the file
        return_value = write_copy_directives_to_file(
            copy_directives=copy_directives,
            file_path=file_path
        )

    except Exception as e:
        return_value.update_sep(
            False,
            "Failed to write duplicate directives with exception: {}".format(e),
        )
    
    return return_value



def write_directives_to_file(swap_directives, copy_directives, output_directory):
    """
    Writes directives to a specified file.

    :return: Result object indicating success or failure.
    :rtype: Result
    """

    return_value = Result()

    try:
        # write swap directives to file
        result_swap = write_swap_directives(swap_directives, output_directory)
        return_value.update(result_swap)

        # write copy directives to file
        result_copy = write_copy_directives(copy_directives=copy_directives, output_directory=output_directory)
        print(result_copy)
        return_value.update(result_copy)

    except Exception as e:
        return_value.update_sep(
            False,
            "Failed to write directives with exception: {}".format(e),
        )
    
    return return_value


def write_families_with_missing_group_codes (family_instances, output_directory):
    """
    Writes a list of families with missing group codes to a specified file.

    :param family_instances: List of family instances with missing group codes.
    :type family_instances: [FamilyTypeDataStorage]
    :param output_directory: Directory where the file will be written.
    :type output_directory: str
    :return: Result object indicating success or failure.
    :rtype: Result
    """
    return_value = Result()

    try:
        # build the file path for the maintain directives
        file_path = os.path.join(output_directory, FAMILIES_WITH_MISSING_GROUP_CODES_FILE_NAME)

        data = []
        for family_instance in family_instances:
            data.append ([
                family_instance.family_name,
            ])

        # write the list to the file
        return_value = write_report_data_as_csv(
            file_name=file_path,
            header=[],  # No header needed for maintain directives
            data= data, 
            quoting= csv.QUOTE_MINIMAL,)

    except Exception as e:
        return_value.update_sep(
            False,
            "Failed to write list of families with missing group codes with exception: {}".format(e),
        )
        
    return return_value



"""
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
Revit API utility functions to set parameter values.
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
"""

#!/usr/bin/python
# -*- coding: utf-8 -*-
# License:
#
#
# Revit Batch Processor Sample Code
#
# BSD License
# Copyright 2023, Jan Christel
# All rights reserved.

# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

# - Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
# - Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
# - Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
#
# This software is provided by the copyright holder "as is" and any express or implied warranties, including, but not limited to, the implied warranties of merchantability and fitness for a particular purpose are disclaimed.
# In no event shall the copyright holder be liable for any direct, indirect, incidental, special, exemplary, or consequential damages (including, but not limited to, procurement of substitute goods or services; loss of use, data, or profits;
# or business interruption) however caused and on any theory of liability, whether in contract, strict liability, or tort (including negligence or otherwise) arising in any way out of the use of this software, even if advised of the possibility of such damage.
#
#

# class used for stats reporting
from duHast.Utilities.Objects import result as res

# import everything from Autodesk Revit DataBase namespace (Revit API)
import Autodesk.Revit.DB as rdb

# utilities
from duHast.Revit.Common import parameter_get_utils as rParaGet
from duHast.Revit.Common import transaction as rTran

# type checker
# from typing import List, Callable

# ----------------------------------------parameters value setters -----------------------------------------------


def set_parameter_value_by_name(
    element,
    parameter_name,  # type: str
    parameter_value_string,
    in_transaction=rTran.in_transaction,
):
    """
    Sets the parameter value by trying to convert the past in string representing the value into the appropriate value type.

    Changing a parameter value requires this action to run inside a transaction.

    :param element: Element of which the parameter value is to be set.
    :type element: Autodesk.Revit.DB.Element
    :param parameter_name: The name of the parameter of which the value is to be set.
    :type parameter_name: str
    :param parameter_value_string: The new parameter value.
    :type parameter_value_string: str
    :param in_transaction: The transaction wrapper function to be used.
    :type in_transaction: func(Autodesk.Revit.DB.Transaction, action(), Autodesk.Revit.DB.Document)

    :raise: Any exception will need to be managed by the function caller.

    """
    return_value = res.Result()

    # get the parameter by name
    para =  element.LookupParameter(parameter_name)

    # check if parameter exists
    if para ==None:
        return_value.update_sep(False, "Parameter not found")
        return return_value
    
    return_value = set_parameter_value(para, parameter_value_string, element.Document, in_transaction)

    return return_value


def set_parameter_value(
    para,
    value_as_string,  # type: str
    doc,
    in_transaction=rTran.in_transaction,
):
    """
    Sets the parameter value by trying to convert the past in string representing the value into the appropriate value type.

    Changing a parameter value requires this action to run inside a transaction.

    :param para: Parameter of which the value is to be set.
    :type para: Autodesk.Revit.DB.Parameter
    :param value_as_string: The new parameter value.
    :type value_as_string: str
    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document
    :param in_transaction: The transaction wrapper function to be used.
    :type in_transaction: func(Autodesk.Revit.DB.Transaction, action(), Autodesk.Revit.DB.Document)

    :raise: Any exception will need to be managed by the function caller.

    ToDo: This needs updating for Revit 2022+ to take into account changes in Revit API: Forge Parameters

    :return:
        Result class instance.

        - Set parameter status (bool) returned in result.status. False if an exception occurred, otherwise True.
        - Result.message property updated in format: Changed parameter value of type x ['parameter name'] : 'old value' to: 'new value'.

        On exception:

        - Set parameter.status (bool) will be False.
        - Set parameter.message will contain the exception message.

    :rtype: :class:`.Result`
    """

    return_value = res.Result()
    old_value = rParaGet.get_parameter_value(para)
    transaction_name = "Update to parameter value"
    # different parameter storage types will require different actions due to value type past in is a string which will need converting
    # first before applied to the parameter
    if para.StorageType == rdb.StorageType.ElementId:
        new_id = rdb.ElementId(int(value_as_string))

        # changing parameter value is required to run inside a transaction
        def action():
            # set up a result instance to be returned to caller with transaction outcome
            action_return_value = res.Result()
            try:
                para.Set(new_id)
                action_return_value.append_message (
                    "Changed parameter value of type Id.[ {} ] from: {} to: {}".format(
                        para.Definition.Name, old_value, value_as_string
                    )
                )
            except Exception as e:
                action_return_value.update_sep(
                    False, "Failed with exception: {}".format(e)
                )
            return action_return_value
        
        # check if a transaction wrapper is required
        if in_transaction:
            transaction = rdb.Transaction(doc, transaction_name)
            return_value = in_transaction(transaction, action)
        else:
            return_value = action() # no transaction wrapper used
    elif para.StorageType == rdb.StorageType.Double:
        # THIS IS THE KEY:  Use SetValueString instead of Set.  Set requires your data to be in
        # whatever internal units of measure Revit uses. SetValueString expects your value to
        # be in whatever the current DisplayUnitType (units of measure) the document is set to
        # for the UnitType associated with the parameter.
        #
        # So SetValueString is basically how the Revit GUI works.
        def action():
            action_return_value = res.Result()
            try:
                para.SetValueString(value_as_string)
                action_return_value.append_message ("Changed parameter value of type double.[ {} ] from: {} to: {}".format(
                    para.Definition.Name, old_value, value_as_string
                ))
            except Exception as e:
                action_return_value.update_sep(
                    False, "Failed with exception: {}".format(e)
                )
            return action_return_value
        
        # check if a transaction wrapper is required
        if in_transaction:
            transaction = rdb.Transaction(doc, transaction_name)
            return_value = in_transaction(transaction, action)
        else:
            return_value = action() # no transaction wrapper used
        
    elif para.StorageType == rdb.StorageType.Integer:

        def action():
            action_return_value = res.Result()
            try:
                para.Set(int(value_as_string))
                action_return_value.append_message("Changed parameter value of type integer.[ {} ] from: {} to: {}".format(
                    para.Definition.Name, old_value, value_as_string
                ))
            except Exception as e:
                action_return_value.update_sep(
                    False, "Failed with exception: {}".format(e)
                )
            return action_return_value
        
        # check if a transaction wrapper is required
        if in_transaction:
            transaction = rdb.Transaction(doc, transaction_name)
            return_value = in_transaction(transaction, action)
        else:
            return_value = action() # no transaction wrapper used
    elif para.StorageType == rdb.StorageType.String:

        def action():
            action_return_value = res.Result()
            try:
                para.Set(value_as_string)
                action_return_value.append_message ("Changed parameter value of type string.[ {} ] from: {} to: {}".format(
                    para.Definition.Name, old_value, value_as_string
                ))
            except Exception as e:
                action_return_value.update_sep(
                    False, "Failed with exception: {}".format(e)
                )
            return action_return_value
        
        # check if a transaction wrapper is required
        if in_transaction:
            transaction = rdb.Transaction(doc, transaction_name)
            return_value = in_transaction(transaction, action)
        else:
            return_value = action() # no transaction wrapper used
    else:
        # dead end
        return_value.update_sep(
            False,
            "Dont know what to do with this storage type: {}".format(para.StorageType),
        )
    return return_value


def set_built_in_parameter_value(
    doc,
    element,
    built_in_parameter_def,
    value_as_string,  # type: str
    parameter_value_setter=set_parameter_value,
):
    """
    Sets the built-in parameter value by trying to convert the past in string representing the value into the appropriate value type.

    Changing a parameter value requires this action to run inside a transaction.

    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document
    :param element: Element to which the built-in parameter belongs.
    :type element: Autodesk.Revit.DB.Element
    :param built_in_parameter_def: The parameters built-in definition of which the value is to be returned.
    :type built_in_parameter_def: Autodesk.Revit.DB.Definition
    :param value_as_string: The new parameter value.
    :type value_as_string: str
    :param parameter_value_setter:
        The function which takes the parameter as an argument and changes it's value to.
        The function needs to accept these args: parameter, new parameter value as string, document
    :type parameter_value_setter: function
    :raise: As per value setter method.

    ToDo: This needs updating for Revit 2022+ to take into account changes in Revit API: Forge Parameters

    :return:
        Result class instance.

        - Set parameter status (bool) returned in result.status. False if an exception occurred, or parameter does not exist on element, otherwise True.
        - Result.message property updated in format: Changed parameter value of type x ['parameter name'] : 'old value' to: 'new value'.

        On exception:

        - Set parameter.status (bool) will be False.
        - Set parameter.message will contain the exception message.

    :rtype: :class:`.Result`
    """

    return_value = res.Result()

    para = element.get_Parameter(built_in_parameter_def)
    if para != None:
        return_value = parameter_value_setter(para, value_as_string, doc)
    else:
        return_value.update_sep(False, "Parameter not found")
    return return_value


def set_parameter_value_simple(
    para,
    value_as_string,  # type: str
):
    """
    Sets the parameter value by trying to convert the past in string representing the value into the appropriate value type.

    Changing a parameter value requires this action to run inside a transaction.

    :param para: Parameter of which the value is to be set.
    :type para: Autodesk.Revit.DB.Parameter
    :param value_as_string: The new parameter value.
    :type value_as_string: str

    :raise: Any exception will need to be managed by the function caller.


    :return:
        Result class instance.

        - Set parameter status (bool) returned in result.status. False if an exception occurred, otherwise True.
        - Result.message property updated in format: Changed parameter value of type x ['parameter name'] : 'old value' to: 'new value'.

        On exception:

        - Set parameter.status (bool) will be False.
        - Set parameter.message will contain the exception message.

    :rtype: :class:`.Result`
    """

    return_value = res.Result()
    old_value = rParaGet.get_parameter_value(para)
    # different parameter storage types will require different actions due to value type past in is a string which will need converting
    # first before applied to the parameter
    if para.StorageType == rdb.StorageType.ElementId:
        new_id = rdb.ElementId(int(value_as_string))
        para.Set(new_id)
        return_value.append_message (
            "Changed parameter value of type Id.[ {} ] from: {} to: {}".format(
                para.Definition.Name, old_value, value_as_string
            )
        )
    elif para.StorageType == rdb.StorageType.Double:
        # THIS IS THE KEY:  Use SetValueString instead of Set.  Set requires your data to be in
        # whatever internal units of measure Revit uses. SetValueString expects your value to
        # be in whatever the current DisplayUnitType (units of measure) the document is set to
        # for the UnitType associated with the parameter.
        #
        # So SetValueString is basically how the Revit GUI works.
        para.SetValueString(value_as_string)
        return_value.append_message (
            "Changed parameter value of type double.[ {} ] from: {} to: {}".format(
                para.Definition.Name, old_value, value_as_string
            )
        )
    elif para.StorageType == rdb.StorageType.Integer:
        para.Set(int(value_as_string))
        return_value.append_message (
            "Changed parameter value of type integer.[ {} ] from: {} to: {}".format(
                para.Definition.Name, old_value, value_as_string
            )
        )
    elif para.StorageType == rdb.StorageType.String:
        para.Set(value_as_string)
        return_value.append_message (
            "Changed parameter value of type string.[ {} ] from: {} to: {}".format(
                para.Definition.Name, old_value, value_as_string
            )
        )
    else:
        # dead end
        return_value.update_sep(
            False,
            "Dont know what to do with this storage type: {}".format(para.StorageType),
        )
    return return_value


def set_parameter_without_transaction_wrapper_by_name(
    element, parameter_name, parameter_value
):
    """
    Sets the parameter value by trying to convert the past in string representing the value into the appropriate value type.

    Changing a parameter value requires this action to run inside a transaction.

    :param element: Element of which the parameter value is to be set.
    :type element: Autodesk.Revit.DB.Element
    :param parameter_name: The name of the parameter of which the value is to be set.
    :type parameter_name: str
    :param parameter_value: The new parameter value.
    :type parameter_value: any

    :return:
        Result class instance.

        - Set parameter status (bool) returned in result.status. False if an exception occurred, otherwise True.
        - Result.message property updated in format: Changed parameter value of type x ['parameter name'] : 'old value' to: 'new value'.

        On exception:

        - Set parameter.status (bool) will be False.
        - Set parameter.message will contain the exception message.
    """

    return_value = res.Result()

    if not isinstance(parameter_value, str):
        parameter_value = str(parameter_value)

    para = element.LookupParameter(parameter_name)

    if para != None:

        # check if the parameter value is a string, if not convert it to a string
        if (isinstance(parameter_value, str)) == False:
            parameter_value = str(parameter_value)

        return_value = set_parameter_value_simple(
            para=para, value_as_string=parameter_value
        )
    else:
        return_value.update_sep(
            False, "Parameter: {} not found.".format(parameter_name)
        )

    return return_value


def set_builtin_parameter_without_transaction_wrapper_by_name(
    element, parameter_definition, parameter_value
):
    """
    Sets the parameter value by trying to convert the past in string representing the value into the appropriate value type.

    Changing a parameter value requires this action to run inside a transaction.

    :param element: Element of which the parameter value is to be set.
    :type element: Autodesk.Revit.DB.Element
    :param parameter_definition: The parameter definition of which the value is to be set.
    :type parameter_definition: Autodesk.Revit.DB.BuiltInParameter
    :param parameter_value: The new parameter value.
    :type parameter_value: any

    :return:
        Result class instance.

        - Set parameter status (bool) returned in result.status. False if an exception occurred, otherwise True.
        - Result.message property updated in format: Changed parameter value of type x ['parameter name'] : 'old value' to: 'new value'.

        On exception:

        - Set parameter.status (bool) will be False.
        - Set parameter.message will contain the exception message.

    """

    return_value = res.Result()

    if not isinstance(parameter_value, str):
        parameter_value = str(parameter_value)

    para = element.get_Parameter(parameter_definition)

    if para != None:
        return_value = set_parameter_value_simple(
            para=para, value_as_string=parameter_value
        )
    else:
        return_value.update_sep(
            False, "Parameter: {} not found.".format(parameter_definition)
        )

    return return_value



"""
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
Module containing functions to write report files from data previously read from file.
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

"""

#
# License:
#
#
# Revit Batch Processor Sample Code
#
# BSD License
# Copyright 2024, Jan Christel
# All rights reserved.

# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

# - Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
# - Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
# - Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
#
# This software is provided by the copyright holder "as is" and any express or implied warranties, including, but not limited to, the implied warranties of merchantability and fitness for a particular purpose are disclaimed.
# In no event shall the copyright holder be liable for any direct, indirect, incidental, special, exemplary, or consequential damages (including, but not limited to, procurement of substitute goods or services; loss of use, data, or profits;
# or business interruption) however caused and on any theory of liability, whether in contract, strict liability, or tort (including negligence or otherwise) arising in any way out of the use of this software, even if advised of the possibility of such damage.
#
#

import os
from csv import QUOTE_MINIMAL

from duHast.Utilities.Objects.result import Result
from duHast.Utilities.files_csv import write_report_data_as_csv
from duHast.Utilities.directory_io import directory_exists


# import report names and identifiers
from duHast.Revit.Family.Data.Objects.family_base_data_processor_defaults import (
    DATA_TYPE_PROCESSOR as data_type_family_base_processor,
)
from duHast.Revit.Family.Data.Objects.family_base_data_processor_defaults import (
    DATA_REPORT_NAME as data_type_family_base_report_name,
)

from duHast.Revit.Categories.Data.Objects.category_data_processor_defaults import (
    DATA_TYPE_PROCESSOR as data_type_category_processor,
)
from duHast.Revit.Categories.Data.Objects.category_data_processor_defaults import (
    DATA_REPORT_NAME as data_type_category_report_name,
)

from duHast.Revit.LinePattern.Data.Objects.line_pattern_data_processor_defaults import (
    DATA_TYPE_PROCESSOR as data_type_line_pattern_processor,
)
from duHast.Revit.LinePattern.Data.Objects.line_pattern_data_processor_defaults import (
    DATA_REPORT_NAME as data_type_line_pattern_report_name,
)

from duHast.Revit.SharedParameters.Data.Objects.shared_parameter_data_processor_defaults import (
    DATA_TYPE_PROCESSOR as data_type_shared_parameter_processor,
)
from duHast.Revit.SharedParameters.Data.Objects.shared_parameter_data_processor_defaults import (
    DATA_REPORT_NAME as data_type_shared_parameter_report_name,
)

from duHast.Revit.Warnings.Data.Objects.warnings_data_processor_defaults import (
    DATA_TYPE_PROCESSOR as data_type_warnings_processor,
)
from duHast.Revit.Warnings.Data.Objects.warnings_data_processor_defaults import (
    DATA_REPORT_NAME as data_type_warnings_report_name,
)


# contains the report name depending on data type
REPORT_NAME_BY_DATA_TYPE = {
    data_type_family_base_processor: data_type_family_base_report_name,
    data_type_category_processor: data_type_category_report_name,
    data_type_line_pattern_processor: data_type_line_pattern_report_name,
    data_type_shared_parameter_processor: data_type_shared_parameter_report_name,
    data_type_warnings_processor: data_type_warnings_report_name,
}


def get_storage_data(family_data):
    """
    Retrieves the storage data as list of strings sorted by storage type.

    :param family_data: List of family instances.
    :type family_data: [:class:`.FamilyDataFamily`]

    :return: A dictionary where the key is the storage data type, and value is a nested list of lists of strings representing the storage data.
    :rtype: {str:[[str]]}
    """

    # loop over all family instances and get the storage data and combine into single dictionary
    # where key is the data type and values is a list of all storage instances of that data type

    return_value = {}
    for family_instance in family_data:
        data_dic = family_instance.get_all_storage_data_as_strings
        for key, item in data_dic.items():
            if key in return_value:
                return_value[key].extend(item)
            else:
                return_value[key] = item
    return return_value


def get_storage_headers(family_data):
    """
    Gets the storage property (report) headers

    :param family_data: List of family instances.
    :type family_data: [:class:`.FamilyDataFamily`]

    :return: A dictionary where the key is the storage data type, and value is a nested list of lists of strings representing the storage data property names.
    :rtype: {str:[[str]]}
    """

    return_value = {}
    for family_instance in family_data:
        data_dic = family_instance.get_all_storage_headers_as_strings
        for key, item in data_dic.items():
            if key not in return_value:
                return_value[key] = item
    return return_value


def write_data_from_families_to_files(family_data, directory_path):
    """
    Writes family data retrieved from report files back to file.

    :param family_data: List of family instances
    :type family_data: [:class:`.FamilyDataFamily`]
    :param directory_path: Fully qualified directory path to write report files to.
    :type directory_path: str
    :raises TypeError: family_data needs to be a list.
    :raises TypeError: directory_path needs to be a string.
    :raises ValueError: Directory: xyz does not exist.
    :raises ValueError: Failed to get any storage data from family data past in.

    :return:
        Result class instance.

        - result.status: Write report files status returned in result.status. False if an exception occurred, otherwise True.
        - result.message: Will contain the fully qualified file path of each of report files written.
        - result.result: Will be a fully qualified file path of each file written.

        On exception

        - Reload.status (bool) will be False
        - Reload.message will contain the exception message

    :rtype: :class:`.Result`
    """

    return_value = Result()

    try:
        # tests first
        if isinstance(family_data, list) == False:
            raise TypeError(
                "Family_data needs to be a list. Got {} instead.".format(
                    type(family_data)
                )
            )

        if isinstance(directory_path, str) == False:
            raise TypeError(
                "Directory_path needs to be a string. Got {} instead".format(
                    type(directory_path)
                )
            )

        if directory_exists(directory_path=directory_path) == False:
            raise ValueError("Directory: {} does not exist.".format(directory_path))

        # get storage data
        storage_data_dic = get_storage_data(family_data=family_data)

        # get storage file headers
        storage_headers_dic = get_storage_headers(family_data=family_data)

        # need to have at least one entry:
        if len(storage_data_dic) < 1:
            raise ValueError("Failed to get any storage data from family data past in.")

        for key, item in storage_data_dic.items():
            # write to file
            if key in REPORT_NAME_BY_DATA_TYPE:
                # build output file name
                full_file_name = os.path.join(
                    directory_path, REPORT_NAME_BY_DATA_TYPE[key], ".csv"
                )
                try:
                    # get file header
                    header_data = []
                    if key in storage_headers_dic:
                        header_data = storage_headers_dic[key]

                    write_result = write_report_data_as_csv(
                        file_name=full_file_name, 
                        header=header_data, 
                        data=item,
                        enforce_ascii=True,
                        encoding="utf-8",
                        bom=None,
                        quoting=QUOTE_MINIMAL,
                    )
                    if write_result.status is False:
                        raise ValueError(write_result.message)
                    
                    return_value.append_message(
                        "Successfully wrote data type: {} report to: {}".format(
                            key, full_file_name
                        )
                    )
                    return_value.result.append(full_file_name)
                except Exception as e:
                    return_value.update_sep(
                        False,
                        "Failed to write report for data type: {} to file: {}".format(
                            key, full_file_name
                        ),
                    )
            else:
                return_value.update_sep(
                    False, "Data type: {} has no report file name assigned".format(key)
                )

    except Exception as e:
        return_value.update_sep(
            False,
            "Failed to write family storage data to file with exception: {}".format(e),
        )

    return return_value



"""
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
A module with helper function around family types data extraction using Revit xml export functions.
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

Supports 2 methods of data extraction:

- from family file on disk
- from family element instance in document

"""

#
# License:
#
#
# Revit Batch Processor Sample Code
#
# BSD License
# Copyright 2024, Jan Christel
# All rights reserved.

# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

# - Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
# - Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
# - Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
#
# This software is provided by the copyright holder "as is" and any express or implied warranties, including, but not limited to, the implied warranties of merchantability and fitness for a particular purpose are disclaimed.
# In no event shall the copyright holder be liable for any direct, indirect, incidental, special, exemplary, or consequential damages (including, but not limited to, procurement of substitute goods or services; loss of use, data, or profits;
# or business interruption) however caused and on any theory of liability, whether in contract, strict liability, or tort (including negligence or otherwise) arising in any way out of the use of this software, even if advised of the possibility of such damage.
#
#
#


import os

from duHast.Utilities.Objects.result import Result
from duHast.Utilities.files_io import (
    get_file_name_without_ext,
    get_directory_path_from_file_path,
)
from duHast.Utilities.files_xml import read_xml_file
from duHast.Revit.Family.Utility.xml_family_type_reader import read_xml_into_storage
from duHast.Revit.Family.Utility.xml_create_atom_exports import (
    write_data_to_temp_xml_file_and_read_it_back,
    write_data_to_xml_file_and_read_it_back,
)


def get_type_data_via_XML_from_family_file(
    application, family_name, family_path, use_temporary_file=True
):
    """
    Get the family type data from the family document using the XML extraction method.
    This can be used to extract the type data from a family document within a Revit session but without opening the family in Revit.

    :param application: The Revit application object.
    :type application: Autodesk.Revit.ApplicationServices.Application
    :param family_name: The name of the family.
    :type family_name: str
    :param family_path: The path of the family file.
    :type family_path: str
    :param use_temporary_file: Whether to use a temporary file for the XML data.
    :type use_temporary_file: bool

    :return: A result object with .result containing a single FamilyTypeDataStorageManager object. (or empty if failed)
    :rtype: Result
    """

    return_value = Result()
    try:
        # set up action to write xml data
        def action(temp_path_xml):
            action_return_value = Result()
            try:
                # Save XML file to temporary location
                # this is a method of the application object and does not require the family to be open...
                application.ExtractPartAtomFromFamilyFile(family_path, temp_path_xml)
                action_return_value.update_sep(True, "Wrote data to XML file.")
            except Exception as e:
                action_return_value.update_sep(
                    False, "Failed to write XML data: {}".format(e)
                )
            return action_return_value

        doc_xml_result = Result()

        if use_temporary_file:
            # Write the data to an XML file and read it back
            doc_xml_result = write_data_to_temp_xml_file_and_read_it_back(action)
        else:
            dir_out = get_directory_path_from_file_path(family_path)
            family_name = get_file_name_without_ext(family_path)
            return_value.append_message(
                "Writing XML data to file: {}".format(
                    os.path.join(dir_out, family_name + ".xml")
                )
            )
            # Write the data to an XML file and read it back
            doc_xml_result = write_data_to_xml_file_and_read_it_back(
                action, os.path.join(dir_out, family_name + ".xml")
            )

        return_value.update(doc_xml_result)

        # check if an xml document was created
        if doc_xml_result.status is False:
            return return_value

        # read the xml data into the storage object
        type_data = read_xml_into_storage(
            doc_xml_result.result, family_name, family_path
        )

        # store list in return object ( clear any previous results )
        return_value.result = [type_data]
    except Exception as e:
        return_value.update_sep(False, "{}".format(e))

    return return_value


def get_family_type_data_from_library(xml_files_in_libraries, progress_callback=None):
    """
    Get the family type data from xml files in the library.
    XML where created by Revit API PartAtomExport method.

    :param xml_files_in_libraries: list of xml files to search for type data
    :type xml_files_in_libraries: [:class:`FileItem`]
    :param progress_callback: progress callback object
    :type progress_callback: :class:`ProgressBase`

    :return:
        Result class instance.

        - result.status: XML conversion status will be returned in result.status. False if an exception occurred, otherwise True.
        - result.message will contain which xml file was read and converted into family type data.
        - result.result will be [:class:`FamilyTypeDataStorageManager`]

        On exception

        - Reload.status (bool) will be False
        - Reload.message will contain the exception message
        - Reload.result will be an empty list

    :rtype: :class:`.Result`
    """

    return_value = Result()
    type_data = []
    try:

        # set progress counter
        counter = 1
        max_value_xml = len(xml_files_in_libraries)

        # update progress
        if progress_callback:
            progress_callback.update(counter, max_value_xml)

        # Sort by file name to ensure consistent order when name is displayed in progress
        sorted_file_paths = sorted(xml_files_in_libraries, key=lambda x: os.path.basename(x.name))

        # get the type data from the library
        for xml_file in sorted_file_paths:

            # get a file name without extension to report progress
            file_name_progress = get_file_name_without_ext(xml_file.name)
            # update progress
            if progress_callback:
                progress_callback.update(counter, max_value_xml, file_name_progress)

            # read xml file
            xml_doc_status = read_xml_file(xml_file.name)
            if xml_doc_status == False:
                return_value.update_sep(
                    False,
                    "Failed to read xml file: {} with exception: {}".format(
                        xml_file.name, xml_doc_status.message
                    ),
                )
                # update progress
                counter = counter + 1
                continue
            else:
                return_value.append_message("Read xml file: {}".format(xml_file.name))

            # get the xml document
            xml_doc = xml_doc_status.result
            if xml_doc is None:
                return_value.update_sep(
                    False, "Failed to read xml file: {}".format(xml_file.name)
                )
                # update progress
                counter = counter + 1
                continue
            else:
                return_value.append_message("Retrieved xml document object")

            # build the family path (required for xml data)
            # assume the xml file is in the same directory as the family file
            fam_name = get_file_name_without_ext(xml_file.name)
            fam_directory = get_directory_path_from_file_path(xml_file.name)
            fam_path = os.path.join(fam_directory, fam_name + ".rfa")

            # load xml data into storage
            return_value.append_message("loading family: {}".format(fam_name))
            xml_data_family = read_xml_into_storage(xml_doc, fam_name, fam_path)

            # add storage to global list
            type_data.append(xml_data_family)

            # update progress
            counter = counter + 1

            # check for user cancel
            if progress_callback != None:
                if progress_callback.is_cancelled():
                    return_value.append_message("User cancelled!")
                    break

    except Exception as e:
        return_value.update_sep(
            False, "Failed to gather family data with exception: {}".format(e)
        )

    # store data to be returned
    return_value.result = type_data

    return return_value



"""
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
This module contains a number of helper functions relating to Revit roofs. 
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
"""
#
# License:
#
#
# Revit Batch Processor Sample Code
#
# BSD License
# Copyright 2023, Jan Christel
# All rights reserved.

# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

# - Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
# - Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
# - Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
#
# This software is provided by the copyright holder "as is" and any express or implied warranties, including, but not limited to, the implied warranties of merchantability and fitness for a particular purpose are disclaimed.
# In no event shall the copyright holder be liable for any direct, indirect, incidental, special, exemplary, or consequential damages (including, but not limited to, procurement of substitute goods or services; loss of use, data, or profits;
# or business interruption) however caused and on any theory of liability, whether in contract, strict liability, or tort (including negligence or otherwise) arising in any way out of the use of this software, even if advised of the possibility of such damage.
#
#
#

import clr
import System

# import Autodesk
import Autodesk.Revit.DB as rdb

from duHast.Revit.Common import common as com
from duHast.Revit.Family import family_utils as rFam
from duHast.Revit.Roofs.Utility import RevitRoofsFilter as rRoofFilter

# --------------------------------------------- utility functions ------------------


def get_all_roof_types_by_category(doc):
    """
    Gets a filtered element collector of all roof types in the model.

    - Basic Roof
    - In place families or loaded families
    - sloped glazing

    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document

    :return: A filtered element collector containing roof types.
    :rtype: Autodesk.Revit.DB.FilteredElementCollector
    """

    collector = rRoofFilter._get_all_roof_types_by_category(doc)
    return collector


def get_all_roof_types_by_class(doc):
    """
    Gets a filtered element collector of all Roof types in the model:

    - Basic Roof
    - sloped glazing

    Since this is based of class roof it will therefore not return any in place family types!

    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document

    :return: A filtered element collector containing roof types.
    :rtype: Autodesk.Revit.DB.FilteredElementCollector
    """

    collector = rRoofFilter._get_roof_types_by_class(doc)
    return collector


# -------------------------------- none in place Roof types -------------------------------------------------------


def get_all_roof_instances_by_category(doc):
    """
    Gets all Roof elements placed in model...ignores in place families (to be confirmed!)

    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document

    :return: A filtered element collector containing roof instances.
    :rtype: Autodesk.Revit.DB.FilteredElementCollector
    """

    return (
        rdb.FilteredElementCollector(doc)
        .OfCategory(rdb.BuiltInCategory.OST_Roofs)
        .WhereElementIsNotElementType()
    )


def get_all_roof_instances_by_class(doc):
    """
    Gets all Roof elements placed in model...ignores roof soffits(???)

    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document

    :return: A filtered element collector containing roof instances.
    :rtype: Autodesk.Revit.DB.FilteredElementCollector
    """

    return (
        rdb.FilteredElementCollector(doc)
        .OfClass(rdb.Roof)
        .WhereElementIsNotElementType()
    )


def get_all_roof_type_ids_by_category(doc):
    """
    Gets all Roof element type ids available in model.

    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document

    :return: List of element ids of roof types.
    :rtype: List Autodesk.Revit.DB.ElementId
    """

    ids = []
    col_cat = get_all_roof_types_by_category(doc)
    ids = com.get_ids_from_element_collector(col_cat)
    return ids


def get_all_roof_type_ids_by_class(doc):
    """
    Gets all Roof element type ids available in model.

    :param doc: _description_
    :type doc: _type_

    :return: List of element ids of roof types.
    :rtype: List Autodesk.Revit.DB.ElementId
    """

    ids = []
    col_class = get_all_roof_types_by_class(doc)
    ids = com.get_ids_from_element_collector(col_class)
    return ids


# -------------------------------- In place Roof types -------------------------------------------------------


def get_in_place_roof_family_instances(doc):
    """
    Gets all instances of in place families of category roof in the model.

    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document

    :return: A filtered element collector containing roof family instances.
    :rtype: Autodesk.Revit.DB.FilteredElementCollector
    """

    filter = rdb.ElementCategoryFilter(rdb.BuiltInCategory.OST_Roofs)
    return (
        rdb.FilteredElementCollector(doc)
        .OfClass(rdb.FamilyInstance)
        .WherePasses(filter)
    )


def get_all_in_place_roof_type_ids(doc):
    """
    Gets type ids off all available in place families of category roof in the model.

    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document

    :return: List of element ids of in place roof types.
    :rtype: List Autodesk.Revit.DB.ElementId
    """

    ids = rFam.get_all_in_place_type_ids_in_model_of_category(
        doc, rdb.BuiltInCategory.OST_Roofs
    )
    return ids



"""
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
Revit families parameter helper functions.
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
"""

#
# License:
#
#
# Revit Batch Processor Sample Code
#
# BSD License
# Copyright 2023, Jan Christel
# All rights reserved.

# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

# - Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
# - Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
# - Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
#
# This software is provided by the copyright holder "as is" and any express or implied warranties, including, but not limited to, the implied warranties of merchantability and fitness for a particular purpose are disclaimed.
# In no event shall the copyright holder be liable for any direct, indirect, incidental, special, exemplary, or consequential damages (including, but not limited to, procurement of substitute goods or services; loss of use, data, or profits;
# or business interruption) however caused and on any theory of liability, whether in contract, strict liability, or tort (including negligence or otherwise) arising in any way out of the use of this software, even if advised of the possibility of such damage.
#
#
#

import clr
import System

# import common library
# utility functions for most commonly used Revit API tasks
from duHast.Revit.Common import transaction as rTran

# class used for stats reporting
from duHast.Utilities.Objects import result as res

# import Autodesk Revit DataBase namespace
from Autodesk.Revit.DB import (
    BuiltInParameterGroup,
    StorageType,
    Transaction,
)


def set_family_parameter_value_by_storage_type(param_w, manager, value):
    """
    Set parameter value based on the storage type of the parameter.

    :param param_w: The parameter to set the value for.
    :type param_w: Parameter
    :param manager: The family manager.
    :type manager: FamilyManager
    :param value: The value to set the parameter to.
    :type value: str

    :return: The result of the operation.
    :rtype: res.Result
    """
    return_value = res.Result()
    try:
        if param_w.StorageType == StorageType.Double:
            # THIS IS THE KEY:  Use SetValueString instead of Set.  Set requires your data to be in
            # whatever internal units of measure Revit uses. SetValueString expects your value to
            # be in whatever the current DisplayUnitType (units of measure) the document is set to
            # for the UnitType associated with the parameter.
            #
            # So SetValueString is basically how the Revit GUI works.
            manager.SetValueString(param_w, value)
            return_value.update_sep(
                True, "Updated {} to value: {}".format(param_w.Definition.Name, value)
            )
        elif param_w.StorageType == StorageType.ElementId:
            manager.Set(param_w, value)
            return_value.update_sep(
                True, "Updated {} to value: {}".format(param_w.Definition.Name, value)
            )
        elif param_w.StorageType == StorageType.Integer:
            manager.Set(param_w, value)
            return_value.update_sep(
                True, "Updated {} to value: {}".format(param_w.Definition.Name, value)
            )
        elif param_w.StorageType == StorageType.String:
            manager.Set(param_w, value)
            return_value.update_sep(
                True, "Updated {} to value: {}".format(param_w.Definition.Name, value)
            )
        else:
            return_value.update_sep(False, "Parameter storage type is not supported!")

    except Exception as e:
        return_value.update_sep(
            False, "failed to set parameter value with exception: {}".format(e)
        )
    return return_value


def set_family_parameter_value(doc, manager, fam_para, value):
    """
    Set parameter value.

    :param doc: The Revit document.
    :type doc: rdb.Document
    :param manager: The family manager.
    :type manager: FamilyManager
    :param fam_para: The parameter to set the value for.
    :type fam_para: Parameter
    :param value: The value to set the parameter to.
    :type value: str

    :return: The result of the operation.
    :rtype: res.Result
    """

    # get the parameter
    param_w = manager.get_Parameter(fam_para.Definition.Name)

    # set-up action to be executed in transaction
    def action():
        action_return_value = res.Result()
        try:
            # attempt to change parameter value
            action_return_value = set_family_parameter_value_by_storage_type(
                param_w, manager, value
            )
        except Exception as e:
            action_return_value.status = False
            action_return_value.append_message (
                "{} : Failed to set parameter value: with exception: {}".format(
                    fam_para.Definition.Name, e
                )
            )
        return action_return_value

    transaction = Transaction(doc, "Setting parameter value")
    return_value = rTran.in_transaction(transaction, action)
    return return_value


def set_parameter_formula(doc, manager, fam_para, formula):
    """
    Set parameter value as a formula.

    :param doc: The Revit document.
    :type doc: rdb.Document
    :param manager: The family manager.
    :type manager: FamilyManager
    :param fam_para: The parameter to set the formula for.
    :type fam_para: rdb.Parameter
    :param formula: The formula to set.
    :type formula: str

    :return: The result of the operation.
    :rtype: res.Result
    """

    def action():
        action_return_value = res.Result()
        try:
            # set parameter formula
            manager.SetFormula(fam_para, formula)
            action_return_value.append_message (
                "{} : parameter formulas successfully set.".format(
                    fam_para.Definition.Name
                )
            )
            action_return_value.result.append(fam_para)
        except Exception as e:
            action_return_value.update_sep(
                False,
                "{} : Failed to set parameter formula: {} with exception: {}".format(
                    fam_para.Definition.Name, formula, e
                ),
            )
        return action_return_value

    transaction = Transaction(doc, "Setting parameter formula")
    return_value = rTran.in_transaction(transaction, action)
    return return_value


def create_family_parameter(
    doc, parameter_name, parameter_group, parameter_type, is_instance
):
    """
    Changes a shared family parameter to a standard family parameter.

    Note: will need updating to support ForgeType fro Revit 2022 onwards

    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document
    :param parameterName: The shared parameter name.
    :type parameterName: str
    :param prefix: Revit requires the new parameter to have a different name to the shard parameter, therefore a prefix to the name is applied, defaults to '_'
    :type prefix: str, optional

    :return:
        Result class instance.

        - Parameter change status returned in result.status. False if an exception occurred, otherwise True.
        - result.message will contain the name of the shared parameter and the new family parameter name.
        - result.status will contain the new family parameter.

        On exception (handled by optimizer itself!):

        - result.status (bool) will be False.
        - result.message will contain generic exception message.
        - result.status will be an empty list

    :rtype: :class:`.Result`
    """

    return_value = res.Result()


    # import Autodesk Revit DataBase namespace
    # ParametertType was removed in revit 2024...need workaround
    from Autodesk.Revit.DB import ParameterType

    # do some type checking
    if not isinstance(parameter_name, str):
        return_value.update_sep(
            False,
            "Parameter name must be a string. Got: {}".format(type(parameter_name)),
        )
        return return_value

    if not isinstance(parameter_group, BuiltInParameterGroup):
        return_value.update_sep(
            False,
            "Parameter group must be a BuiltInParameterGroup. Got: {}".format(
                type(parameter_group)
            ),
        )
        return return_value

    if not isinstance(parameter_type, ParameterType):
        return_value.update_sep(
            False,
            "Parameter type must be a ParameterType. Got: {}".format(
                type(parameter_type)
            ),
        )
        return return_value

    if not isinstance(is_instance, bool):
        return_value.update_sep(
            False, "Is instance must be a bool. Got: {}".format(type(is_instance))
        )
        return return_value

    # get the family manager
    manager = doc.FamilyManager
    # set up an action to be executed in a transaction
    def action():
        action_return_value = res.Result()
        try:

            parameter_new = manager.AddParameter(
                parameter_name, parameter_group, parameter_type, is_instance
            )

            action_return_value.update_sep(
                True, "Added parameter: {} to family.".format(parameter_name)
            )
            action_return_value.result.append(parameter_new)
        except Exception as e:
            action_return_value.update_sep(
                False,
                "Failed to add parameter: {} with exception: {}".format(
                    parameter_name, e
                ),
            )
        return action_return_value

    # set up a transaction and execute it
    transaction = Transaction(doc, "Add family parameter: {}".format(parameter_name))
    return_value = rTran.in_transaction(transaction, action)

    return return_value

def associate_parameter_with_other_parameter_on_nested_family_instance(
    doc, 
    nested_family_instance, 
    target_parameter_name, 
    source_parameter_name,
    transaction_manager = rTran.in_transaction,
):
    """
    Associate a nested family instance parameter ( can be type or instance) with a host family parameter.

    :param doc: The Revit document.
    :type doc: Document
    :param nested_family_instance: The nested family instance.
    :type nested_family_instance: FamilyInstance
    :param target_parameter_name: The target parameter name.
    :type target_parameter_name: str
    :param source_parameter_name: The source parameter name.
    :type source_parameter_name: str

    :return: The result of the operation.
    :rtype: res.Result
    """

    return_value = res.Result()
    # get the family manager
    manager = doc.FamilyManager
    # get the nested family placed instance parameters
    nested_instance_parameters = nested_family_instance.GetOrderedParameters()
    # get the hot family parameters
    host_family_parameters = manager.GetParameters()

    # set up some flags to check if all went well
    found_source_parameter = False
    found_target_parameter = False

    # loop over the nested family instance parameters and check for matching parameter name
    for nested_instance_parameter in nested_instance_parameters:

        # check if match found for target parameter name
        if nested_instance_parameter.Definition.Name == target_parameter_name:
            target_parameter = nested_instance_parameter
            found_target_parameter = True

            # loop over the host family parameters and check for matching source parameter name
            for host_family_parameter in host_family_parameters:

                # check if match found for source parameter name
                if host_family_parameter.Definition.Name == source_parameter_name:
                    source_parameter = host_family_parameter
                    found_source_parameter = True

                    # set up an action to be executed in a transaction
                    def action():
                        action_return_value = res.Result()
                        try:
                            manager.AssociateElementParameterToFamilyParameter(
                                target_parameter, source_parameter
                            )
                            action_return_value.update_sep(
                                True,
                                "Associated parameter: {} with parameter: {}".format(
                                    target_parameter_name, source_parameter_name
                                ),
                            )
                        except Exception as e:
                            action_return_value.update_sep(
                                False,
                                "Failed to associate parameter: {} with parameter: {} with exception: {}".format(
                                    target_parameter_name, source_parameter_name, e
                                ),
                            )
                        return action_return_value
                    
                    # execute the action
                    if transaction_manager:
                        # if in transaction, execute the action in the transaction
                        transaction_manager = rTran.in_transaction
                        transaction = Transaction(
                            doc,
                            "Associate parameter: {} with {}".format(
                                target_parameter_name, source_parameter_name
                            ),
                        )
                        return_value = rTran.in_transaction(transaction, action)
                    else:
                        # if not in transaction, execute the action without a transaction
                        return_value = action()
                    break

            # break out of the loop if source parameter found
            if found_source_parameter:
                break
            else:
                # give user feedback if source parameter not found
                return_value.update_sep(
                    False,
                    "Failed to find source parameter: {}".format(source_parameter_name),
                )
                break

    # check if target parameter found
    if not found_target_parameter:
        return_value.update_sep(
            False, "Failed to find target parameter: {}".format(target_parameter_name)
        )

    return return_value

def filter_parameters_by_formula_driven(parameters, keep_if_formulae_driven):
    """
    Filter the parameters by whether they are determined by a formula or not.

    :param parameters: The list of parameters
    :type parameters: list
    :param keep_if_formulae_driven: True to keep the parameters that are determined by a formula, False otherwise
    :type keep_if_formulae_driven: bool
    :return: The filtered list of parameters
    :rtype: list
    """
    
    return_list = []
    for parameter in parameters:
        if parameter.IsDeterminedByFormula and keep_if_formulae_driven:
            return_list.append(parameter)
        elif not parameter.IsDeterminedByFormula and not keep_if_formulae_driven:
            return_list.append(parameter)
    return return_list


def get_family_type_parameters(doc):
    """
    Get the type parameters of the family.

    :param doc: The family document
    :type doc: Document

    :return:
       List of type parameters in the family document.
       if an exception occurred or the document is not a family document, None is returned.

    :rtype: :class:`.Result`
    """

    type_parameters = []

    # check if this is a family document
    if not doc.IsFamilyDocument:
        return None
    
    # Assume 'doc' is your current Document object
    family_manager = doc.FamilyManager

    try:
        # Iterate over all parameters in the family and get the type parameters
        for parameter in family_manager.Parameters:
            # Check if it's a type parameter and has a formula
            if not parameter.IsInstance:
                type_parameters.append(parameter)
        
    except:
        return None

    return type_parameters



def get_all_parameters_where_parameter_is_in_formula(doc, target_parameter):
    

    """
    Get all parameters where the target parameter is used in a formula.
    
    Args:
        doc: The Revit document.
        target_parameter: The parameter to check in formulas.
        
    Returns:
        A list of parameters that use the target parameter in their formulas.
    """
    
    parameters = []
    
    for param in doc.FamilyManager.Parameters:
        if param.IsFormula and target_parameter.Name in param.GetFormula():
            parameters.append(param)
    
    return parameters


# License:
#
#
# Revit Batch Processor Sample Code
#
# BSD License
# Copyright 2025, Jan Christel
# All rights reserved.

# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

# - Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
# - Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
# - Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
#
# This software is provided by the copyright holder "as is" and any express or implied warranties, including, but not limited to, the implied warranties of merchantability and fitness for a particular purpose are disclaimed.
# In no event shall the copyright holder be liable for any direct, indirect, incidental, special, exemplary, or consequential damages (including, but not limited to, procurement of substitute goods or services; loss of use, data, or profits;
# or business interruption) however caused and on any theory of liability, whether in contract, strict liability, or tort (including negligence or otherwise) arising in any way out of the use of this software, even if advised of the possibility of such damage.
#
#


import os

# import workloader utils
from duHast.UI import workloader as wl
from duHast.Utilities.Objects.result import Result

# import file list methods
from duHast.UI import file_list as fl

from duHast.Utilities.files_io import (
    is_back_up_file,
)

def write_task_lists(family_directory, task_list_directory, number_of_task_lists):
    

    return_value = Result()
    try:

        get_data = fl.bucket_to_task_list_file_system
        
        # get revit files in input dir
        revit_files_unfiltered = fl.get_revit_files(
            family_directory, ".rfa"
        )
        
        revit_files = []
        for revit_file in revit_files_unfiltered:
            # remove any back up files from selection
            if is_back_up_file(os.path.basename(revit_file.name)) == False:
                if (
                    len(os.path.dirname(os.path.abspath(revit_file.name))) < 248
                    and len(revit_file.name) < 260
                ):
                    revit_files.append(revit_file)
                else:
                    return_value.append_message(
                        "Max path length violation: {}".format(revit_file.name)
                    )
                    return_value.append_message("File has been removed from selection!")


        # build bucket list
        buckets = wl.distribute_workload(
            number_of_task_lists, revit_files, fl.get_file_size
        )
        # write out file lists
        counter = 0
        for bucket in buckets:
            file_name = os.path.join(
                task_list_directory, "Tasklist_" + str(counter) + ".txt"
            )
            status_write = fl.write_revit_task_file(file_name, bucket, get_data)
            return_value.update(status_write)
            counter += 1
    except Exception as e:
        return_value.update_sep(
            False,
            "Failed to write task lists with exception: {}".format(e),
        )
    
    return return_value


"""
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
Helper functions relating to xml files.
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
"""

#
# License:
#
#
# Revit Batch Processor Sample Code
#
# BSD License
# Copyright 2025, Jan Christel
# All rights reserved.

# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

# - Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
# - Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
# - Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
#
# This software is provided by the copyright holder "as is" and any express or implied warranties, including, but not limited to, the implied warranties of merchantability and fitness for a particular purpose are disclaimed.
# In no event shall the copyright holder be liable for any direct, indirect, incidental, special, exemplary, or consequential damages (including, but not limited to, procurement of substitute goods or services; loss of use, data, or profits;
# or business interruption) however caused and on any theory of liability, whether in contract, strict liability, or tort (including negligence or otherwise) arising in any way out of the use of this software, even if advised of the possibility of such damage.
#
#
#


import clr

clr.AddReference("System.Xml")
from System.Xml import XmlDocument
import xml.etree.ElementTree as ET

from duHast.Utilities.files_base_read import read_non_column_based_text_file
from duHast.Utilities.Objects.result import Result
from duHast.UI.file_list import get_revit_files


def read_xml_file(file_path):
    """
    Reads an XML file and returns the content as a string.
    :param file_path: The path to the XML file.

    :return:
        Result class instance.

        - result.status (bool) True if file was read without an exception, otherwise False.
        - result.message contains log messages.
        - result.result will the XML Document instance.

        On exception:

        - result.status (bool) will be False.
        - result.message will contain exception message.
    :rtype: :class:`.Result`
    """

    return_value = Result()

    read_result = read_non_column_based_text_file(file_path)
    return_value.update(read_result)

    if read_result.status is False:
        return return_value

    try:
        # Load the XML content
        doc_xml = XmlDocument()
        doc_xml.LoadXml(read_result.result)
        return_value.result = doc_xml

    except Exception as e:
        return_value.update_sep(False, "Error while reading the XML file: {}".format(e))

    return return_value


def get_xml_files_from_directory(directory):
    """
    Gets all XML files from a directory.
    :param directory: The directory to search for XML files.
    :type directory: str

    :return: A list of XML files.
    :rtype: [:class:`.FileItem`]
    """

    xml_files = []
    try:
        xml_files = get_revit_files(directory, "*.xml")
    except Exception:
        pass
    return xml_files


def get_all_xml_files_from_directories(process_directories):
    """
    Get all xml files from the directories

    :param process_directories: list of directories to search for xml files
    :type process_directories: list

    :return: list of xml files found
    :rtype: [:class:`FileItem`]
    """

    files_found = []
    try:
        # get all xml files from the directory
        for directory in process_directories:
            files = get_xml_files_from_directory(directory)
            files_found = files_found + files
    except Exception as e:
        raise Exception("Failed to gather xml files with exception: {}".format(e))
    return files_found


def get_xml_element_value(path, tag_name, default=None, value_type=str):
    if not path:
        raise ValueError("Path to XML file cannot be None or empty.")
    if not tag_name:
        raise ValueError("Tag name cannot be None or empty.")
    tree = ET.parse(path)
    root = tree.getroot()
    element = None
    for elem in root.iter():
        if elem.tag == tag_name:
            element = elem

    if element is not None and element.text is not None:
        try:
            return value_type(element.text.strip())
        except ValueError:
            return default
    return default



# License:
#
#
# Revit Batch Processor Sample Code
#
# BSD License
# Copyright 2025, Jan Christel
# All rights reserved.

# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

# - Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
# - Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
# - Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
#
# This software is provided by the copyright holder "as is" and any express or implied warranties, including, but not limited to, the implied warranties of merchantability and fitness for a particular purpose are disclaimed.
# In no event shall the copyright holder be liable for any direct, indirect, incidental, special, exemplary, or consequential damages (including, but not limited to, procurement of substitute goods or services; loss of use, data, or profits;
# or business interruption) however caused and on any theory of liability, whether in contract, strict liability, or tort (including negligence or otherwise) arising in any way out of the use of this software, even if advised of the possibility of such damage.
#
#

from duHast.Utilities.files_csv import read_csv_file

def clean_code(code):
    """
    Cleans the grouping code by removing any sub codes after a dot.
    """
    if code is None:
        return None
    
    # remove any sub codes after a dot
    if '.' in code:
        code = code.split('.')[0]
    
    return code


def convert_to_file_name_code(code):
    """
    Converts the grouping code to a file name safe format.
    """

    # return code as is for now
    return code

    # if code is None:
    #     return None
    
    # # replace "-" with "_"
    # code = code.replace('-', '_')
    
    # return code


def load_group_code_description(file_path):
    """
    Loads the grouping code and description from a CSV file.
    
    :param file_path: The path to the CSV file containing the grouping codes and descriptions.
    :return: A dictionary mapping cleaned grouping codes to their descriptions.
    """

    code_description_map = {}
    try:
        read_result = read_csv_file(file_path=file_path)
        
        if not read_result.status:
            return None
        
        code_description_map = {}
        
        # ignore header row
        if len(read_result.result) == 0:
            return  None
        
        # iterate over the rows and create a map of code to description
        # ignore header row
        for i in range(1, len(read_result.result)):
            row = read_result.result[i]
            if len(row) < 2:
                continue
            code = row[0]
            family_code = row[1]
            description = row[2]
            
            # add to map
            code_description_map[code] = (family_code, description)
    
    except Exception as e:
        print("Error loading grouping codes from file {}: {}".format(file_path, e))
        return None
    

    return code_description_map


"""
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
Family type data class.
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
"""

#
# License:
#
#
# Revit Batch Processor Sample Code
#
# BSD License
# Copyright 2023, Jan Christel
# All rights reserved.

# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

# - Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
# - Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
# - Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
#
# This software is provided by the copyright holder "as is" and any express or implied warranties, including, but not limited to, the implied warranties of merchantability and fitness for a particular purpose are disclaimed.
# In no event shall the copyright holder be liable for any direct, indirect, incidental, special, exemplary, or consequential damages (including, but not limited to, procurement of substitute goods or services; loss of use, data, or profits;
# or business interruption) however caused and on any theory of liability, whether in contract, strict liability, or tort (including negligence or otherwise) arising in any way out of the use of this software, even if advised of the possibility of such damage.
#
#
#


from duHast.Revit.Family.Data.Objects import ifamily_data as IFamData
from duHast.Revit.Family.Data.Objects.family_type_data_storage import (
    FamilyTypeDataStorage,
)
from duHast.Revit.Family.Data.Objects.family_type_parameter_data_storage import FamilyTypeParameterDataStorage
from duHast.Revit.Family.Data.Objects.family_type_data_processor_defaults import (
    NESTING_SEPARATOR,
)

from duHast.Revit.Family.family_functions import get_name_and_category_to_family_dict
from duHast.Revit.Family.family_types_model_get_data_from_xml import get_type_data_via_XML_from_family_object
# import Autodesk
from  Autodesk.Revit.DB import Element

# data dictionary key values specific to this class
CATEGORY_NAME = "categoryName"


class FamilyTypeData(IFamData.IFamilyData):
    def __init__(self, root_path=None, root_category_path=None):
        """
        Class constructor

        :param rootPath: The path of the nested family in a tree: rootFamilyName :: nestedFamilyNameOne :: nestedFamilyTwo\
            This includes the actual family name as the last node.
        :type rootPath: str
        :param rootCategoryPath: The path of the family category in a tree: rootCategoryName :: nestedCategoryNameOne :: nestedCategoryTwo\
            This includes the actual category name as the last node.
        :type rootCategoryPath: str
        """

        # store data type  in base class
        super(FamilyTypeData, self).__init__(
            root_path=root_path,
            root_category_path=root_category_path,
        )
        # super(CategoryData, self).__init__(rootPath, dataType)

        if root_category_path != None:
            category_chunks = root_category_path.split(NESTING_SEPARATOR)
            self.category = category_chunks[-1]
        else:
            self.category = "unknown"

        self.saved_file_name = ""


    def process(self, doc, session_id):
        """
        Collects all base data from the document and stores it in the class property .data

        :param doc: Current family document
        :type doc: Autodesk.Revit.DB.Document
        """

        # make sure to get a value for the file path which is not empty if the document has not been saved
        if doc.PathName != "":
            self.saved_file_name = doc.PathName

        # get all nested families and export to xml
        families = get_name_and_category_to_family_dict(doc)
        
        for family_name, family in families.items():
            
            
            fam_name = self.root_path
            
            
            # get the type data and make sure to pass in root category path and root name path
            try:
                # add the family processed to the path
                fam_name = Element.Name.GetValue(family)
                # strip .rfa of name
                if fam_name.lower().endswith(".rfa"):
                    fam_name = family_name[:-4]
                fam_root_path = "{}{}{}".format(self.root_path,NESTING_SEPARATOR ,fam_name)
            except Exception as e:
                #print("Failed to get family name: {}".format(e))
                fam_root_path = "{}{}{}".format(self.root_path, NESTING_SEPARATOR, e)

            fam_cat_name = self.root_category_path


            try:
                # get the category of the family to be processed
                fam_cat_name = family.FamilyCategory.Name
                fam_root_category_path = "{}{}{}".format(self.root_category_path , NESTING_SEPARATOR , fam_cat_name)
            except Exception as e:
                #print("Failed to get family category: {}".format(e))
                fam_root_category_path = "{}{}{}".format(self.root_category_path, NESTING_SEPARATOR, e)

            type_data_result = get_type_data_via_XML_from_family_object(family, fam_root_path, fam_root_category_path)
            if type_data_result.status and len(type_data_result.result) > 0:
                
                # in the moment data contains a list of storage objects rather than a single storage manager object
                storage_manager = type_data_result.result[0]
                for type_storage in storage_manager.family_type_data_storage:
                    # we actually want the nested parameter storage here to get a single entry per
                    # family, family type and parameter
                    for  parameter_storage in type_storage.parameters:
                        self.add_data(parameter_storage)
                    
                
    def get_data(self):
        return self.data

    def add_data(self, storage_instance):
        if isinstance(storage_instance, FamilyTypeParameterDataStorage):
            self.data.append(storage_instance)
        else:
            raise ValueError(
                "storage instance must be an instance of FamilyTypeParameterDataStorage got {}".format(type(storage_instance))
            )



# License:
#
#
# Revit Batch Processor Sample Code
#
# BSD License
# Copyright 2025, Jan Christel
# All rights reserved.

# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

# - Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
# - Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
# - Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
#
# This software is provided by the copyright holder "as is" and any express or implied warranties, including, but not limited to, the implied warranties of merchantability and fitness for a particular purpose are disclaimed.
# In no event shall the copyright holder be liable for any direct, indirect, incidental, special, exemplary, or consequential damages (including, but not limited to, procurement of substitute goods or services; loss of use, data, or profits;
# or business interruption) however caused and on any theory of liability, whether in contract, strict liability, or tort (including negligence or otherwise) arising in any way out of the use of this software, even if advised of the possibility of such damage.
#
#
#
import os

from duHast.Utilities.Objects.result import Result
from duHast.Utilities.files_io import get_file_name_without_ext

from duHast.Revit.Family.Utility.family_swap_instances_by_type_utils import get_swap_directives
from duHast.Revit.Family.family_swap_instances_of_types import swap_family_instances_of_types
from duHast.Revit.Family.LibraryCleanUp.Utility.family_load import load_families_required_for_swapping

from duHast.Revit.Purge.purge_unused_e_transmit import purge_unused_e_transmit
from duHast.Revit.Common.file_io import save_as_family

from duHast.Revit.RBP.Objects.ProgressRBPConsole import ProgressRBPConsole
from duHast.pyRevit.Objects.ProgressPyRevit import ProgressPyRevit


DEBUG = True



def in_package_family(doc, library_path, output_directory, output):
    """
    
    """

    return_value = Result()

    # read maintain list and delete unused types in family
    # export family types (?)

    # find families which need swapping load them and swap them out
    # save family

    try:

        # make sure this really is a family document
        if not doc.IsFamilyDocument:
            return_value.update_sep(False, "The document is not a family document.")
            return return_value

        # load swap directives
        swap_directives_result = get_swap_directives(directory_path=library_path)

        if not swap_directives_result.status:
            return_value.update_sep(False, "Failed to read swap directives: {}".format(swap_directives_result.message))
            output("Failed to read swap directives: {}".format(swap_directives_result.message))
            return return_value
        
        output("Swap directives loaded successfully. Found {} directives.".format(len(swap_directives_result.result)))
        swap_directives = swap_directives_result.result

        #output("Loading families required for swapping...")
        #output("Swap directives: {}".format(swap_directives))
        #output("Library path: {}".format(library_path))
        # load families required for swapping
        load_result = load_families_required_for_swapping(doc, swap_directives, library_path)
        

        if not load_result.status:
            return_value.update_sep(False, "Failed to load families required for swapping: {}".format(load_result.message))
            output("Failed to load families required for swapping: {}".format(load_result.message))
            return return_value
        
        #output("Families successfully loaded for swapping: {}".format(len(load_result.result)))
        #output("Families loaded: {}".format([fam.Name for fam in load_result.result]))

        # swap out families if necessary
        if( len(load_result.result) == 0):
            return_value.update_sep(True, "No families needing swapping in file")
            output("No families needing swapping in file")
            #return return_value
        else:
            # swap away
            swap_result = swap_family_instances_of_types(doc, library_path, progress_callback=None)

            # print logs
            output("Swapped families with status: {}".format(swap_result.status))

            # check what came back
            if not swap_result.status:
                return_value.update_sep(False, "Failed to swap families: {}".format(swap_result.message))
                output("Failed to swap families: {}".format(swap_result.message))
                return return_value
       
        
        # if we get here... do a purge unused and than save that family document
        purge_action_result = purge_unused_e_transmit(doc)

        if not purge_action_result.status:
            return_value.update_sep(False, "Failed to purge unused elements: {}".format(purge_action_result.message))
            output("Failed to purge unused elements: {}".format(purge_action_result.message))
            return return_value
        
        output("Purged unused elements successfully.")

        file_name_without_ext = get_file_name_without_ext(doc.PathName)
        
        revit_fam_file_path = os.path.join( output_directory, file_name_without_ext + ".rfa")
        output("Saving family to: {}".format(revit_fam_file_path))

        # save the family document
        save_result = save_as_family(
            doc=doc,
            target_directory_path= output_directory,
            current_full_file_name= doc.PathName,
            name_data=[[file_name_without_ext, file_name_without_ext]],
            file_extension= ".rfa",
            compact_file=True)
        
        output("Save result: {}".format(save_result.status))


    except Exception as e:
        return_value.update_sep(False, "An error occurred: {}".format(str(e)))
        output("An error occurred: {}".format(str(e)))
        return return_value
    return return_value


"""
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
Family category data processor class.
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
"""

#
# License:
#
#
# Revit Batch Processor Sample Code
#
# BSD License
# Copyright 2023, Jan Christel
# All rights reserved.

# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

# - Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
# - Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
# - Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
#
# This software is provided by the copyright holder "as is" and any express or implied warranties, including, but not limited to, the implied warranties of merchantability and fitness for a particular purpose are disclaimed.
# In no event shall the copyright holder be liable for any direct, indirect, incidental, special, exemplary, or consequential damages (including, but not limited to, procurement of substitute goods or services; loss of use, data, or profits;
# or business interruption) however caused and on any theory of liability, whether in contract, strict liability, or tort (including negligence or otherwise) arising in any way out of the use of this software, even if advised of the possibility of such damage.
#
#
#

from duHast.Revit.Family.Data.Objects.ifamily_processor import IFamilyProcessor
from duHast.Revit.Categories.Data.Objects import category_data as rCatData
from duHast.Revit.Categories.Data.Objects.category_data_processor_defaults import (
    DATA_TYPE_PROCESSOR as data_type_category_processor,
)
from duHast.Revit.Family.Data.Objects.family_base_data_processor_defaults import (
    NESTING_SEPARATOR,
)
from duHast.Revit.Family.Data.Objects import ifamily_data as IFamData
from duHast.Revit.Family.Data.Objects.ifamily_data_storage import IFamilyDataStorage
from duHast.Utilities.Objects import result as res


class CategoryProcessor(IFamilyProcessor):

    data_type = data_type_category_processor

    def __init__(self, pre_actions=None, post_actions=None):
        """
        Class constructor.
        """

        # store data type  in base class
        super(CategoryProcessor, self).__init__(
            pre_actions=pre_actions,
            post_actions=[self._post_action_update_used_subcategories],
            data_type=CategoryProcessor.data_type,
        )

        # self.data = []
        # self.dataType = 'Category'

        # list of corner cases when it comes to category checking: imports in families or Reference planes ... ( english language specific!! )
        self.category_check_corner_cases = ["Imports in Families", "Reference Planes"]

        # self.preActions = preActions
        # set default post action to updated categories used in root processor with any categories found in nested
        # families
        # self.postActions = [self._postActionUpdateUsedSubcategories]
        # add any other post actions
        if post_actions != None:
            for p_action in post_actions:
                self.post_actions.append(p_action)

    def process(self, doc, root_path, root_category_path):
        """
        Calls processor instance with the document and root path provided and adds processor instance to class property .data

        :param doc: Current family document.
        :type doc: Autodesk.Revit.DB.Document
        :param root_path: The path of the nested family in a tree: rootFamilyName :: nestedFamilyNameOne :: nestedFamilyTwo\
            This includes the actual family name as the last node.
        :type root_path: str
        :param root_category_path: The category path of the nested family in a tree: rootFamilyCategory :: nestedFamilyOneCategory :: nestedFamilyTwoCategory\
            This includes the actual family category as the last node.
        :type root_category_path: str
        """

        dummy = rCatData.CategoryData(root_path, root_category_path)
        dummy.process(doc)
        self.data.append(dummy)

    # --------------------------------------------- post action ----------------------------------------------------------

    def _is_sub_category_present(
        self, root_family_data, nested_family_sub_category_storage
    ):
        """
        Check if sub category is present in root family data.

        :param root_family_data: List of IFamilyData objects.
        :type root_family_data: list
        :param nested_family_sub_category_storage: IFamilyDataStorage object.
        :type nested_family_sub_category_storage: IFamilyDataStorage
        :return: IFamilyData object.
        :rtype: IFamilyData
        """

        # check what came is a list
        if isinstance(root_family_data, list) == False:
            raise ValueError(
                "Root family data must be a list but is type: {}".format(
                    type(root_family_data)
                )
            )
        if isinstance(nested_family_sub_category_storage, IFamilyDataStorage) == False:
            raise ValueError(
                "Nested family sub category storage must be an instance of IFamilyDataStorage but is type: {}".format(
                    type(nested_family_sub_category_storage)
                )
            )

        match = None
        # check whether sub category is present
        for root_fam in root_family_data:
            if isinstance(root_fam, IFamData.IFamilyData) == False:
                raise ValueError(
                    "Root family data must be a list of IFamilyData objects but is type: {}".format(
                        type(root_fam)
                    )
                )
            storage_instances = root_fam.get_data()
            for storage_root in storage_instances:
                if (
                    storage_root.category_name
                    == nested_family_sub_category_storage.category_name
                    and storage_root.sub_category_name
                    == nested_family_sub_category_storage.sub_category_name
                ):
                    match = root_fam
                    break
        return match

    def _update_root_family_data(
        self, root_family_data, nested_families_sub_categories
    ):
        """
        Update root family data with used subcategories from nested families.

        :param root_family_data: List of IFamilyData objects.
        :type root_family_data: list
        :param nested_families_sub_categories: List of IFamilyDataStorage objects.
        :type nested_families_sub_categories: list
        """

        # check what came is a list
        if isinstance(root_family_data, list) == False:
            raise ValueError(
                "Root family data must be a list but is type: {}".format(
                    type(root_family_data)
                )
            )
        if isinstance(nested_families_sub_categories, list) == False:
            raise ValueError(
                "Nested families sub categories must be a list but is type: {}".format(
                    type(nested_families_sub_categories)
                )
            )

        # loop over nested family subcategory storage data
        for nested_sub_category_storage in nested_families_sub_categories:

            if isinstance(nested_sub_category_storage, IFamilyDataStorage) == False:
                raise ValueError(
                    "Nested family sub category must be an instance of IFamilyDataStorage but is type: {}".format(
                        type(nested_sub_category_storage)
                    )
                )

            # check if sub category is already in root family
            matching_root_fam_category_data = self._is_sub_category_present(
                root_family_data, nested_sub_category_storage
            )
            if matching_root_fam_category_data != None:
                root_storage_all = matching_root_fam_category_data.get_root_storage()

                # some data instances might have more than one root storage instance to represent multiple categories present in the family
                for root_storage in root_storage_all:
                    if (
                        nested_sub_category_storage.category_name
                        == root_storage.category_name
                    ) and (
                        nested_sub_category_storage.sub_category_name
                        == root_storage.sub_category_name
                    ):
                        # update used by list
                        if (
                            nested_sub_category_storage.family_name
                            not in root_storage.used_by
                        ):
                            root_storage.update_usage(nested_sub_category_storage)
            else:
                pass
                # nothing to do if that category has not been reported to start off with
                # this category could, for example, belong to the section marker family present in most 3d families

    def _get_used_subcategories(self, data_instances):
        """
        Get used subcategories from nested families data as list of IFamilyDataStorage objects.

        :param data_instances: List of IFamilyData objects.
        :type data_instances: list
        :return: List of IFamilyDataStorage objects.
        :rtype: list
        """

        # data_instances should be a list of IFamilyData objects
        if isinstance(data_instances, list) == False:
            raise ValueError(
                "Data must be a list but is type: {}".format(type(data_instances))
            )

        used_subcategories = []
        for d in data_instances:

            # check what is being processed...should be an IFamilyData object
            if isinstance(d, IFamData.IFamilyData) == False:
                raise ValueError(
                    "Data must be an instance of IFamilyData but is type: {}".format(
                        type(d)
                    )
                )

            # TODO: check if the root category path is correct
            # should I get the first or last entry?? Might need to be the first?
            # i'm not entirely sure why I'm checking this here...
            # is this so I only delete sub categories in the root family which are not in use any where in the nested families of the same category?
            # i.e Furniture.OverHead assumes the host family is of Category Furniture and I am checking whether the sub category OverHead is in use in any
            # of the nested families as well as the host family?

            # The report (storage data ) also contains two fields: Category and SubCategory. Category in this case can well be different to the category of the family!

            # loop over storage of the data instance
            data_storage_instances = d.get_data()
            for storage in data_storage_instances:
                if storage.use_counter > 0:
                    # get the family category
                    category_path = storage.root_category_path.split(NESTING_SEPARATOR)
                    # which is the last entry in the root category path
                    category_storage = category_path[len(category_path) - 1]
                    # select only items which either belong to the category of the family or
                    # are corner cases like imports in families or Reference planes ... ( english language specific!! )
                    if (
                        category_storage == storage.category_name
                        or storage.category_name in self.category_check_corner_cases
                    ):
                        used_subcategories.append(storage)
        return used_subcategories

    def _post_action_update_used_subcategories(self, doc):
        """
        Post action to update used subcategories data in the root family data object.

        :param doc: Current family document.
        :type doc: Autodesk.Revit.DB.Document
        :return: Result object.
        :rtype: Result
        """

        return_value = res.Result()
        try:
            # find all subcategories of nested families
            nested_family_data = self._find_nested_families_data()
            # get used sub categories from nested data
            nested_family_used_sub_categories = self._get_used_subcategories(
                data_instances=nested_family_data
            )
            # update root family data only
            root_family_data = self._find_root_family_data()
            # update root processor data as required
            self._update_root_family_data(
                root_family_data, nested_family_used_sub_categories
            )
            return_value.update_sep(
                True, "Post Action Update subcategories data successful completed."
            )
        except Exception as e:
            return_value.update_sep(
                False,
                "Post Action Update subcategories data failed with exception: {}".format(
                    e
                ),
            )
        return return_value



"""
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
Helper functions relating to system processes.
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
"""
#
# License:
#
#
# Revit Batch Processor Sample Code
#
# BSD License
# Copyright 2023, Jan Christel
# All rights reserved.

# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

# - Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
# - Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
# - Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
#
# This software is provided by the copyright holder "as is" and any express or implied warranties, including, but not limited to, the implied warranties of merchantability and fitness for a particular purpose are disclaimed.
# In no event shall the copyright holder be liable for any direct, indirect, incidental, special, exemplary, or consequential damages (including, but not limited to, procurement of substitute goods or services; loss of use, data, or profits;
# or business interruption) however caused and on any theory of liability, whether in contract, strict liability, or tort (including negligence or otherwise) arising in any way out of the use of this software, even if advised of the possibility of such damage.
#
#
#
import subprocess
import os, signal

from duHast.Utilities import date_stamps as dateStamp


def get_all_running_processes():
    """
    Retrieves a list of all currently running processes.

    :return: a list in format: [[HandleCount, Name, Priority, ProcessId, ThreadCount, WorkingSetSize]]
    :rtype: list of list
    """

    # traverse the software list
    Data = subprocess.check_output(["wmic", "process", "list", "brief"])
    a = str(Data)
    #  arrange the string
    processUnfiltered = []
    # list of all processes running... process ID is at index 3
    processFiltered = []
    try:
        processUnfiltered = a.split("\r\r\n")
        counter = 0
        columnNumber = 0
        for i in range(len(processUnfiltered)):
            if counter == 0:
                counter = +1
                columnNumber = len(processUnfiltered[i].split())
            else:
                dummyList = processUnfiltered[i].split()
                if len(dummyList) == columnNumber:
                    processFiltered.append(dummyList)
                else:
                    dif = len(dummyList) - columnNumber
                    processName = ""
                    for i in range(dif + 1):
                        processName = processName + " " + dummyList[i + 1]
                    fixedProcessData = []
                    fixedProcessData.append(dummyList[0])
                    fixedProcessData.append(processName)
                    for i in range(dif + 2, len(dummyList), 1):
                        fixedProcessData.append(dummyList[i])
                    processFiltered.append(fixedProcessData)
    except IndexError as e:
        # print ('Got all running processes')
        pass
    return processFiltered


def filter_by_process_name(processNames, processList, returnMatch=True):
    """
    Filters a provided list of processes by process name

    :param processNames: List of names to filter by
    :type processNames: list of str
    :param processList: List of processes running.
    :type processList: [[HandleCount, Name, Priority, ProcessId, ThreadCount, WorkingSetSize]]
    :param returnMatch: If true only matches will be returned, if false any non matches will be returned, defaults to True
    :type returnMatch: bool, optional

    :return: List of processes
    :rtype: [[HandleCount, Name, Priority, ProcessId, ThreadCount, WorkingSetSize]]
    """

    processFilteredByName = []
    for process in processList:
        match = False
        for processName in processNames:
            if process[1] == processName:
                match = True
                break
        if returnMatch == True and match == True:
            processFilteredByName.append(process)
        if returnMatch == False and match == False:
            processFilteredByName.append(process)
    return processFilteredByName


def filter_by_process_ids(processIds, processList, returnMatch=True):
    """
    Filters a provided list of processes by process ids

    :param processIds: List of ids to filter by
    :type processIds: _type_
    :param processList: List of processes running
    :type processList: [[HandleCount, Name, Priority, ProcessId, ThreadCount, WorkingSetSize]]
    :param returnMatch: If true only matches will be returned, if false any non matches will be returned, defaults to True
    :type returnMatch: bool, optional
    :return: List of processes
    :rtype: [[HandleCount, Name, Priority, ProcessId, ThreadCount, WorkingSetSize]]
    """

    processFilteredByName = []
    for process in processList:
        match = False
        for processId in processIds:
            if process[3] == processId:
                match = True
                break
        if returnMatch == True and match == True:
            processFilteredByName.append(process)
        if returnMatch == False and match == False:
            processFilteredByName.append(process)
    return processFilteredByName


def kill_processes(processList):
    """
    Kills all processes in list provided

    :param processList: List of processes to be killed.
    :type processList: [[HandleCount, Name, Priority, ProcessId, ThreadCount, WorkingSetSize]]

    :return: True if all past in processes have been killed, otherwise False.
    :rtype: bool
    """

    status = True
    for process in processList:
        try:
            os.kill(int(process[3]), signal.SIGTERM)
            print(str(process[1]) + " " + str(process[3]) + " killed")
        except Exception as e:
            status = False
            print(e)
    return status


def write_out_process_data(directoryPath, processList, fileName, fileExtension):
    """
    Writes out process data to file

    :param directoryPath: The directory path to where the export is being saved.
    :type directoryPath: str
    :param processList: List of processes to be written to file.
    :type processList: [[HandleCount, Name, Priority, ProcessId, ThreadCount, WorkingSetSize]]
    :param fileName:  The file name under which the export is being saved (excluding file extension).
    :type fileName: str
    :param fileExtension: in format '.extension'
    :type fileExtension: str

    :return: True if the process data was written to file without an exception, otherwise False.
    :rtype: bool
    """

    status = True
    # setup file name
    filePath = (
        directoryPath
        + "\\"
        + dateStamp.get_file_date_stamp(dateStamp.FILE_DATE_STAMP_YYYY_MM_DD_HH_MM_SEC)
        + fileName
        + fileExtension
    )
    try:
        f = open(filePath, "w")
        for p in processList:
            f.writelines(",".join(p))
        f.close()
    except:
        status = False
    return status



# License:
#
#
# Revit Batch Processor Sample Code
#
# BSD License
# Copyright 2025, Jan Christel
# All rights reserved.

# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

# - Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
# - Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
# - Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
#
# This software is provided by the copyright holder "as is" and any express or implied warranties, including, but not limited to, the implied warranties of merchantability and fitness for a particular purpose are disclaimed.
# In no event shall the copyright holder be liable for any direct, indirect, incidental, special, exemplary, or consequential damages (including, but not limited to, procurement of substitute goods or services; loss of use, data, or profits;
# or business interruption) however caused and on any theory of liability, whether in contract, strict liability, or tort (including negligence or otherwise) arising in any way out of the use of this software, even if advised of the possibility of such damage.
#
#

import os

from duHast.Utilities.Objects.result import Result
from duHast.Utilities.files_io import file_exist
from duHast.Revit.Family.family_utils import get_all_loadable_family_ids_through_types
from duHast.Revit.Family.family_functions import get_name_and_category_to_family_dict, get_symbol_names_of_family
from duHast.Revit.Family.Data.Objects.family_base_data_processor_defaults import NESTING_SEPARATOR
from duHast.Revit.Family import family_utils as rFamUtil
from duHast.UI.Objects.ProgressBase import ProgressBase



from Autodesk.Revit.DB import Element

def get_families_to_be_loaded_for_swapping(doc, swap_directives):
    """
    Identifies all families required to be loaded into a project for swapping based on the provided swap directives.

    Will check if target family already exists in the project and if so discard it from the list of families to be loaded.

    :param doc: The Revit document.
    :type doc: Autodesk.Revit.DB.Document
    :param swap_directives: The swap directives containing family names and paths.
    :type swap_directives: list of dict
    :return: List of loaded family symbols.
    :rtype: list of Autodesk.Revit.DB.FamilySymbol
    """
    
    return_value = Result()

    try:
        # get all loaded families:
        family_ids = get_all_loadable_family_ids_through_types(doc)
        # get family name and category dictionary
        family_name_and_category_dict = get_name_and_category_to_family_dict(doc)

        if len(family_ids) > 0:
            return_value.append_message(
                "Found:  {} loadable families in file.".format(len(family_ids))
            )

            for fam_id in family_ids:
                fam = doc.GetElement(fam_id)
                fam_name = Element.Name.GetValue(fam)
                # get the types in the model
                source_family_symbol_names = get_symbol_names_of_family(fam)
                # return_value.append_message(
                #     "Family: {}, Category: {}".format(
                #         fam_name, fam.FamilyCategory.Name
                #     )
                # )
                for source_family_symbol_name in source_family_symbol_names:
                    #return_value.append_message("...Checking symbol: {}".format(source_family_symbol_name))
                    # symbol has match?
                    symbol_has_match=False
                    # loop over swap directives and try to find a match for the family name and category and type name
                    for swap_directive in swap_directives:
                        #return_value.append_message("......Checking swap directive: {}".format(swap_directive.name, swap_directive.category, swap_directive.source_type_name))
                        if swap_directive.name == fam_name and fam.FamilyCategory.Name == swap_directive.category and swap_directive.source_type_name == source_family_symbol_name:
                            #return_value.append_message(".........Found match")
                            # set flag that we found a match
                            symbol_has_match = True
                        
                            # check if target family already exists in the project
                            target_fam_key = "{}{}{}".format(
                                swap_directive.target_family_name, NESTING_SEPARATOR, swap_directive.category
                            )

                            # check if the target family is already loaded
                            if target_fam_key in family_name_and_category_dict:
                                # family already loaded, skip it only if target type is present, otherwise assume a reload is required
                                # get all type names for this family
                                #return_value.append_message(".........Family {} is loaded, checking symbol.".format(swap_directive.target_family_name))
                                
                                # get the target family symbol names
                                target_family = family_name_and_category_dict[target_fam_key]
                                # get the symbol names of the target family
                                target_family_symbol_names = get_symbol_names_of_family(target_family)
                                
                                # check if the target family symbol is present in the family
                                if  swap_directive.target_family_type_name in  target_family_symbol_names:
                                    return_value.append_message(".........Family {} and type {} already loaded, skipping.".format(swap_directive.target_family_name, swap_directive.target_family_type_name))
                                    # family symbol already loaded, skip it
                                    continue
                                else:
                                    pass
                                    #return_value.append_message(".........Type {} not loaded".format(swap_directive.target_family_type_name))
                            else:
                                pass
                                # family not loaded, we need to load it
                                #return_value.append_message(".........Family {} not loaded, marked for loading.".format(swap_directive.target_family_name))
                                
                            # return_value.append_message(
                            #     "...Family {} and type {} needs swapping to family {} and type {}.".format(swap_directive.name,  swap_directive.source_type_name, swap_directive.target_family_name, swap_directive.target_family_type_name)
                            # )

                            # only add if not there already
                            if(swap_directive.target_family_name not in return_value.result):
                                # add family name to result list
                                return_value.result.append(swap_directive.target_family_name)
                                return_value.append_message(
                                    "...Family {} added to load list.".format(swap_directive.target_family_name)
                                )
                            else:
                                return_value.append_message(
                                    "...Family {} already in load list, skipping.".format(swap_directive.target_family_name)
                                )
                            
                            break  # break out of the loop since we found a match
                    
                    if symbol_has_match == False:
                        pass
                        # return_value.append_message(
                        #     "{},{},{}, not found in swap directives.".format(fam_name, fam.FamilyCategory.Name, symbol_name)
                        # )
                        
    except Exception as e:
        return_value.update_sep(
            False,
            "Failed to build load families list: {}".format(e),
        )
    
    return return_value


def load_families_required_for_swapping(doc, swap_directives, library_directory, progress_callback=None):
    """
    Loads families required for swapping into the document.

    :param doc: The Revit document.
    :type doc: Autodesk.Revit.DB.Document
    :param swap_directives: The swap directives containing family names and paths.
    :type swap_directives: list of dict
    :param library_directory: The directory where family files are located.
    :type library_directory: str
    :param progress_callback: Optional progress callback for UI updates.
    :type progress_callback: ProgressBase or None

    :return: List of loaded family symbols.
    :rtype: list of Autodesk.Revit.DB.FamilySymbol
    """
    
    return_value = Result()

    try:

        # check callback class
        if progress_callback and isinstance(progress_callback, ProgressBase) == False:
            raise TypeError(
                "progress_callback needs to be inherited from ProgressBase. Got : {} instead.".format(
                    type(progress_callback)
                )
            )
    
        # get families to be loaded for swapping
        family_load_required_result = get_families_to_be_loaded_for_swapping(doc, swap_directives)

        # DEBUG
        #print(family_load_required_result.message)

        # check if the family load required result is valid
        if family_load_required_result.status == False:
            return_value.update_sep(
                False,
                family_load_required_result.message,
            )
            return return_value
        
        # check if any families are required to be loaded
        if len(family_load_required_result.result) == 0:
            return_value.update_sep(
                True,
                "No families to load required for swapping.",
            )
            return return_value
        
        # progress call back
        callback_counter = 1

        # get the max progress value
        loop_max = len(family_load_required_result.result)

        # load the required families
        for family_name in family_load_required_result.result:

            # update progress
            if progress_callback != None:
                progress_callback.update(callback_counter, loop_max)
            
            # build the family load path
            family_load_path = os.path.join(library_directory, family_name + ".rfa")

            if (file_exist(family_load_path) == False):
                return_value.update_sep(
                    False,
                    "Family file {} does not exist in library directory.".format(family_load_path),
                )
                # skip and move to next family
                continue

            # load families
            result_load = rFamUtil.load_family(doc,family_load_path)
           
            if not result_load.status:
                print("Failed to load family {}: {}".format(family_name, result_load.message))
                return_value.update_sep(
                    False,
                    "Failed to load family {}: {}".format(family_name, result_load.message),
                )
                continue
            
            return_value.append_message("Successfully loaded family: {}".format(family_name))
            return_value.result.append(result_load.result[0])
        
            # update progress
            callback_counter = callback_counter + 1

            # check for progress cancel
            if progress_callback != None:
                if progress_callback.is_cancelled():
                    return_value.append_message("User cancelled!")
                    break
        
    except Exception as e:
        return_value.update_sep(
            False,
            "Failed to load families for swapping: {}".format(e),
        )
    
    return return_value


"""
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
Data bljueprint class for Room Layout sheets.
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

This class has the following properties:

- room size
- room proportion
- number of items in room
- sheets

"""

#
# License:
#
#
# Revit Batch Processor Sample Code
#
# BSD License
# Copyright 2024, Jan Christel
# All rights reserved.

# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

# - Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
# - Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
# - Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
#
# This software is provided by the copyright holder "as is" and any express or implied warranties, including, but not limited to, the implied warranties of merchantability and fitness for a particular purpose are disclaimed.
# In no event shall the copyright holder be liable for any direct, indirect, incidental, special, exemplary, or consequential damages (including, but not limited to, procurement of substitute goods or services; loss of use, data, or profits;
# or business interruption) however caused and on any theory of liability, whether in contract, strict liability, or tort (including negligence or otherwise) arising in any way out of the use of this software, even if advised of the possibility of such damage.
#
#
#

import json
from duHast.Utilities.Objects.base import Base
from duHast.Data.Objects.BluePrints.blueprint_property_names import (
    BlueprintPropertyNames,
)


class RoomLayoutSheet(Base):
    def __init__(self, j=None, **kwargs):
        """
        Class constructor
        """

        # forwards all unused arguments
        # ini super class to allow multi inheritance in children!
        super(RoomLayoutSheet, self).__init__(**kwargs)

        self._room_size = 0.0
        self._room_proportions = 1.0
        self._room_number_of_items = 0

        self._sheets = []

        json_var = None
        # check if any data was past in with constructor!
        if j is not None:
            # check type of data that came in:
            if isinstance(j, str):
                # a string
                json_var = json.loads(j)
            elif isinstance(j, dict):
                # no action required
                json_var = j.copy()
            else:
                raise TypeError(
                    "Argument j supplied must be of type string or type dictionary. Got {} instead.".format(
                        type(j)
                    )
                )

            # attempt to populate from json
            try:
                pass

            except Exception as e:
                raise type(e)(
                    "Node {} failed to initialise with: {}".format(self.data_type, e)
                )

    def __eq__(self, other):
        if not isinstance(other, RoomLayoutSheet):
            return NotImplemented
        return (
            self._room_size == other.set_name
            and self.option_name == other.option_name
            and self.is_primary == other.is_primary
        )

    def __ne__(self, other):
        return not self.__eq__(other)

    def __hash__(self):
        return hash(
            (self._room_size, self._room_proportions, self._room_number_of_items)
        )
